from enum import Enum
from typing import Dict, List, Optional

from pydantic import Field, constr, validator

from ..exceptions import ValidationError
from .basemodels import ProtoModel


class HarmonicType(str, Enum):
    """
    The angular momentum representation of a shell.
    """

    spherical = "spherical"
    cartesian = "cartesian"


class ElectronShell(ProtoModel):
    """
    Information for a single electronic shell
    """

    angular_momentum: List[int] = Field(..., description="Angular momentum for this shell.")
    harmonic_type: HarmonicType = Field(..., description=str(HarmonicType.__doc__))
    exponents: List[float] = Field(..., description="Exponents for this contracted shell.")
    coefficients: List[List[float]] = Field(
        ...,
        description="General contraction coefficients for this shell, individual list components will be the individual segment contraction coefficients.",
    )

    @validator("coefficients")
    def _check_coefficient_length(cls, v, values):
        len_exp = len(values["exponents"])
        for row in v:
            if len(row) != len_exp:
                raise ValueError("The length of coefficients does not match the length of exponents.")

        return v

    @validator("coefficients")
    def _check_general_contraction_or_fused(cls, v, values):
        if len(values["angular_momentum"]) > 1:
            if len(values["angular_momentum"]) != len(v):
                raise ValueError("The length for a fused shell must equal the length of coefficients.")

        return v

    def nfunctions(self) -> int:
        """
        Computes the number of basis functions on this shell.

        Returns
        -------
        int
            The number of basis functions on this shell.
        """

        if self.harmonic_type == "spherical":
            return sum((2 * L + 1) for L in self.angular_momentum)
        else:
            return sum(((L + 1) * (L + 2) // 2) for L in self.angular_momentum)

    def is_contracted(self) -> bool:
        """
        Checks if the shell represents a contracted Gaussian or not.

        Returns
        -------
        bool
            True if the shell is contracted.
        """

        return (len(self.coefficients) != 1) and (len(self.angular_momentum) == 1)


class ECPType(str, Enum):
    """
    The type of the ECP potential.
    """

    scalar = "scalar"
    spinorbit = "spinorbit"


class ECPPotential(ProtoModel):
    """
    Information for a single ECP potential.
    """

    ecp_type: ECPType = Field(..., description=str(ECPType.__doc__))
    angular_momentum: List[int] = Field(..., description="Angular momentum for the ECPs.")
    r_exponents: List[int] = Field(..., description="Exponents of the 'r' term.")
    gaussian_exponents: List[float] = Field(..., description="Exponents of the 'gaussian' term.")
    coefficients: List[List[float]] = Field(
        ...,
        description="General contraction coefficients for this shell, individual list components will be the individual segment contraction coefficients.",
    )

    @validator("gaussian_exponents")
    def _check_gaussian_exponents_length(cls, v, values):
        len_exp = len(values["r_exponents"])
        if len(v) != len_exp:
            raise ValueError("The length of gaussian_exponents does not match the length of `r` exponents.")

        return v

    @validator("coefficients")
    def _check_coefficient_length(cls, v, values):
        len_exp = len(values["r_exponents"])
        for row in v:
            if len(row) != len_exp:
                raise ValueError("The length of coefficients does not match the length of `r` exponents.")

        return v


class BasisCenter(ProtoModel):
    """
    Data for a single atom/center in a basis set.
    """

    electron_shells: List[ElectronShell] = Field(..., description="Electronic shells for this center.")
    ecp_electrons: int = Field(0, description="Number of electrons replace by ECP potentials.")
    ecp_potentials: Optional[List[ECPPotential]] = Field(None, description="ECPs for this center.")


class BasisSet(ProtoModel):
    """
    A quantum chemistry basis description.
    """

    schema_name: constr(strip_whitespace=True, regex="qcschema_basis") = "qcschema_basis"
    schema_version: int = 1

    name: str = Field(..., description="A standard basis name if available (e.g., 'cc-pVDZ').")
    description: Optional[str] = Field(None, description="A brief description of the basis set.")
    center_data: Dict[str, BasisCenter] = Field(..., description="A mapping of all types of centers available.")
    atom_map: List[str] = Field(
        ..., description="Mapping of all centers in the parent molecule to centers in `center_data`."
    )

    nbf: Optional[int] = Field(None, description="The number of basis functions.")

    @validator("atom_map")
    def _check_atom_map(cls, v, values):
        sv = set(v)
        missing = sv - values["center_data"].keys()

        if missing:
            raise ValueError(f"'atom_map' contains unknown keys to 'center_data': {missing}.")

        return v

    @validator("nbf", always=True)
    def _check_nbf(cls, v, values):

        # Bad construction, pass on errors
        try:
            nbf = cls._calculate_nbf(values["atom_map"], values["center_data"])
        except KeyError:
            return v

        if v is None:
            v = nbf
        else:
            if v != nbf:
                raise ValidationError("Calculated nbf does not match supplied nbf.")

        return v

    @classmethod
    def _calculate_nbf(cls, atom_map, center_data) -> int:
        """
        Number of basis functions in the basis set.

        Returns
        -------
        int
            The number of basis functions.
        """

        center_count = {}
        for k, center in center_data.items():
            center_count[k] = sum(x.nfunctions() for x in center.electron_shells)

        ret = 0
        for center in atom_map:
            ret += center_count[center]

        return ret



from typing import Dict

from qcelemental.extras import get_information


def provenance_stamp(routine: str) -> Dict[str, str]:
    """Return dictionary satisfying QCSchema,
    https://github.com/MolSSI/QCSchema/blob/master/qcschema/dev/definitions.py#L23-L41
    with QCElemental's credentials for creator and version. The
    generating routine's name is passed in through `routine`.

    """
    return {"creator": "QCElemental", "version": get_information("version"), "routine": routine}



from copy import deepcopy
from typing import Any, Dict, Union

import numpy as np

from ..exceptions import ValidationError
from ..physical_constants import constants
from ..util import unnp
from .to_string import formula_generator


def to_schema(
    molrec: Dict[str, Any], dtype: Union[str, int], units: str = "Bohr", *, np_out: bool = False, copy: bool = True
) -> Dict[str, Any]:
    """Translate molparse internal Molecule spec into dictionary from other schemas.

    Parameters
    ----------
    molrec : dict
        Psi4 json Molecule spec.
    dtype : {'psi4', 1, 2}
        Molecule schema format.
        ``1`` is https://molssi-qc-schema.readthedocs.io/en/latest/auto_topology.html V1 + #44 + #53
        ``2`` is ``1`` with internal schema_name/version (https://github.com/MolSSI/QCSchema/pull/60)
    units : {'Bohr', 'Angstrom'}
        Units in which to write string. There is not an option to write in
        intrinsic/input units. Some `dtype` may not allow all units.
    np_out : bool, optional
        When `True`, fields originating from geom, elea, elez, elem, mass, real, elbl will be ndarray.
        Use `False` to get a json-able version.
    #return_type : {'json', 'yaml'} Serialization format string to return.

    Returns
    -------
    qcschema : dict
        Dictionary of the `dtype` repr of `molrec`.

    """
    qcschema: Dict = {}

    geom = np.array(molrec["geom"], copy=copy)

    if molrec["units"] == "Bohr" and units == "Bohr":
        pass
    elif molrec["units"] == "Angstrom" and units == "Bohr" and "input_units_to_au" in molrec:
        geom = geom * molrec["input_units_to_au"]
    else:
        geom = geom * constants.conversion_factor(molrec["units"], units)

    nat = geom.shape[0] // 3

    name = molrec.get("name", formula_generator(molrec["elem"]))
    #    tagline = """auto-generated by qcdb from molecule {}""".format(name)

    if dtype == "psi4":
        if units not in ["Angstrom", "Bohr"]:
            raise ValidationError(
                """Psi4 Schema {} allows only 'Bohr'/'Angstrom' coordinates, not {}.""".format(dtype, units)
            )
        qcschema = deepcopy(molrec)
        qcschema["geom"] = geom
        qcschema["units"] = units
        qcschema["name"] = name

    elif dtype in [1, 2]:
        if units != "Bohr":
            raise ValidationError("""QCSchema {} allows only 'Bohr' coordinates, not {}.""".format(dtype, units))

        molecule: Dict = {}
        molecule["validated"] = True
        molecule["symbols"] = np.array(molrec["elem"], copy=copy)
        molecule["geometry"] = geom
        molecule["masses"] = np.array(molrec["mass"], copy=copy)
        molecule["atomic_numbers"] = np.array(molrec["elez"], copy=copy)
        molecule["mass_numbers"] = np.array(molrec["elea"], copy=copy)
        molecule["atom_labels"] = np.array(molrec["elbl"], copy=copy)
        molecule["name"] = name
        if "comment" in molrec:
            molecule["comment"] = molrec["comment"]
        molecule["molecular_charge"] = molrec["molecular_charge"]
        molecule["molecular_multiplicity"] = molrec["molecular_multiplicity"]
        molecule["real"] = np.array(molrec["real"], copy=copy)
        fidx = np.split(np.arange(nat), molrec["fragment_separators"])
        molecule["fragments"] = [fr.tolist() for fr in fidx]
        molecule["fragment_charges"] = np.array(molrec["fragment_charges"]).tolist()
        molecule["fragment_multiplicities"] = np.array(molrec["fragment_multiplicities"]).tolist()
        molecule["fix_com"] = molrec["fix_com"]
        molecule["fix_orientation"] = molrec["fix_orientation"]
        if "fix_symmetry" in molrec:
            molecule["fix_symmetry"] = molrec["fix_symmetry"]
        molecule["provenance"] = deepcopy(molrec["provenance"])
        if "connectivity" in molrec:
            molecule["connectivity"] = deepcopy(molrec["connectivity"])

        if dtype == 1:
            qcschema = {"schema_name": "qcschema_input", "schema_version": 1, "molecule": molecule}
        elif dtype == 2:
            qcschema = molecule
            qcschema.update({"schema_name": "qcschema_molecule", "schema_version": 2})

    else:
        raise ValidationError(
            "Schema dtype not understood, valid options are {{'psi4', 1, 2}}. Found {}.".format(dtype)
        )

    if not np_out:
        qcschema = unnp(qcschema)

    return qcschema

    # if return_type == 'json':
    #    return json.dumps(qcschema)
    # elif return_type == 'yaml':
    #    import yaml
    #    return yaml.dump(qcschema)
    # else:
    #    raise ValidationError("""Return type ({}) not recognized.""".format(return_type))



import re
from enum import Enum, EnumMeta
from textwrap import dedent, indent
from typing import Any

from pydantic import BaseModel, BaseSettings

__all__ = ["auto_gen_docs_on_demand", "get_base_docs", "AutoPydanticDocGenerator"]


class AutoDocError(ValueError):
    """
    Helper error Raised when the autodoc application failed.

    Traps this very specific error and not other ValueErrors
    """


def type_to_string(input_type):
    if type(input_type) is type:
        return input_type.__name__
    else:
        return repr(input_type).replace("typing.", "")


def is_pydantic(test_object):
    try:
        instance = isinstance(test_object, BaseSettings) or isinstance(test_object, BaseModel)
    except TypeError:
        instance = False
    try:
        subclass = issubclass(test_object, BaseSettings) or issubclass(test_object, BaseModel)
    except TypeError:
        subclass = False
    return instance or subclass


def parse_type_str(prop) -> str:
    from pydantic import fields  # Import here to minimize issues

    typing_map = {
        fields.SHAPE_TUPLE: "Tuple",
        fields.SHAPE_SET: "Set",
        fields.SHAPE_LIST: "List",
        fields.SHAPE_SINGLETON: "Union",
    }  # yapf: disable
    if type(prop) is type or prop.__module__ == "typing":
        # True native Python type
        prop_type_str = type_to_string(prop)
    elif issubclass(prop.type_.__class__, Enum) or issubclass(prop.type_.__class__, EnumMeta):
        # Enumerate, have to do the __class__ or issubclass(prop.type_) throws issues later.
        prop_type_str = "{" + ",".join([str(x.value) for x in prop.type_]) + "}"
    elif type(prop.type_) is type and prop.shape == fields.SHAPE_SINGLETON:
        # Native Python type buried in a Field
        prop_type_str = type_to_string(prop.type_)
    elif is_pydantic(prop.type_):
        # Pydantic types
        prop_type_str = f":class:`{prop.type_.__name__}`"
    elif prop.type_.__module__ == "typing":
        # Typing Types
        prop_type_str = ""
        key_field = prop.key_field
        sub_fields = prop.sub_fields
        # Special case of Optional[]
        # if sub_fields is not None and any([sf.sub_fields is type(None) for sf in sub_fields]):
        if sub_fields is not None and any([sf.type_ is type(None) for sf in sub_fields]):
            reconstructed_props = [f for f in sub_fields if f.type_ is not type(None)]
            parsed_types = [parse_type_str(f) for f in reconstructed_props]
            if len(parsed_types) == 1:
                prop_type_str = parsed_types[0]
            else:
                prop_type_str = "Union[" + ", ".join(parsed_types) + "]"
        elif prop.shape == fields.SHAPE_MAPPING:
            prop_type_str = "Dict[" + parse_type_str(key_field) + ", " + parse_type_str(prop.type_) + "]"
        elif sub_fields is not None:
            # Not "optional", but iterable
            prop_type_str = typing_map[prop.shape] + "[" + ", ".join([parse_type_str(sf) for sf in sub_fields]) + "]"
        elif prop.type_ is Any:
            prop_type_str = "Any"
    elif "ConstrainedInt" in prop.type_.__name__:
        prop_type_str = "ConstrainedInt"
    elif "ConstrainedFloat" in prop.type_.__name__:
        prop_type_str = "ConstrainedFloat"
    elif prop.shape in typing_map.keys():
        if prop.sub_fields is None:
            # Single item
            if prop.type_.__module__ == "pydantic.types":
                # A bit of a catch-all
                prop_type_str = prop.type_.__name__
            else:
                prop_type_str = typing_map[prop.shape] + "[" + parse_type_str(prop.type_) + "]"
        else:
            prop_type_str = (
                typing_map[prop.shape] + "[" + ", ".join([parse_type_str(sf) for sf in prop.sub_fields]) + "]"
            )
    else:
        # Finally, with nothing else to do...
        prop_type_str = str(prop)

    return prop_type_str


def doc_formatter(base_docs: str, target_object: BaseModel, allow_failure: bool = True) -> str:
    """
    Generate the docstring for a Pydantic object automatically based on the parameters

    This could use improvement.

    Might be ported to Elemental at some point
    """

    # Convert the None to regex-parsable string
    if base_docs is None:
        doc_edit = ""
    else:
        doc_edit = base_docs

    # Is pydantic and not already formatted
    if is_pydantic(target_object) and not re.search(r"^\s*Parameters\n", doc_edit, re.MULTILINE):
        try:
            # Add the white space
            if not doc_edit.endswith("\n\n"):
                doc_edit += "\n\n"
            # Add Parameters separate
            new_doc = dedent(doc_edit) + "Parameters\n----------\n"
            # Get Pydantic fields
            target_fields = target_object.__fields__
            # Go through each property
            for prop_name, prop in target_fields.items():
                # Handle Type
                prop_type_str = parse_type_str(prop)

                # Handle (optional) description
                prop_desc = prop.field_info.description  # type: ignore

                # Combine in the following format:
                # name : type(, Optional, Default)
                #   description
                first_line = prop_name + " : " + prop_type_str
                if not prop.required and (prop.default is None or is_pydantic(prop.default)):
                    first_line += ", Optional"
                elif prop.default is not None:
                    first_line += f", Default: {prop.default}"
                # Write the prop description
                second_line = "\n" + indent(prop_desc, "    ") if prop_desc is not None else ""
                # Finally, write the detailed doc string
                new_doc += first_line + second_line + "\n"
        except:  # lgtm [py/catch-base-exception]
            if allow_failure:
                new_doc = base_docs
            else:
                raise

    else:
        new_doc = base_docs

    # Assign the new doc string
    return new_doc


class AutoPydanticDocGenerator:
    """
    Dynamic Doc generator, should never be called directly and only though augo_gen_docs_on_demand or as a part of the
    __new__ constructor in a metaclass.
    """

    ALREADY_AUTODOCED_ATTR = "__model_autodoc_applied__"
    AUTODOC_BASE_DOC_REFERENCE_ATTR = "__base_doc__"

    def __init__(self, target: BaseModel, allow_failure: bool = True, always_apply: bool = False):
        # Checks against already instanced and uninstanced classes while avoiding unhahsable type error

        if not always_apply:
            if isinstance(target, BaseModel) or (isinstance(target, type) and issubclass(target, BaseModel)):

                if (
                    hasattr(target, self.ALREADY_AUTODOCED_ATTR)
                    and getattr(target, self.ALREADY_AUTODOCED_ATTR) is True
                ):
                    raise AutoDocError(
                        "Object already has autodoc rules applied to it, cannot re-apply auto documentation"
                        f"without first resetting the __doc__ attribute and setting "
                        f"{self.ALREADY_AUTODOCED_ATTR} to False (or deleting it)"
                    )
            else:
                raise TypeError("Cannot use auto-doc tech on non-BaseModel subclasses")

        self.base_doc = target.__doc__
        setattr(target, self.AUTODOC_BASE_DOC_REFERENCE_ATTR, self.base_doc)
        self.target = target
        setattr(target, self.ALREADY_AUTODOCED_ATTR, True)
        self.allow_failure = allow_failure

    def __get__(self, *args):
        return doc_formatter(self.base_doc, self.target, allow_failure=self.allow_failure)

    def __del__(self):
        try:
            self.target.__doc__ = self.base_doc
            if hasattr(self.target, self.ALREADY_AUTODOCED_ATTR):
                setattr(self.target, self.ALREADY_AUTODOCED_ATTR, False)
        except:  # lgtm [py/catch-base-exception]
            # Corner case where trying to reapply and failing cannot delete the new self mid __init__ since
            # base_doc has not been set.
            pass


def auto_gen_docs_on_demand(
    target: BaseModel, allow_failure: bool = True, ignore_reapply: bool = True, force_reapply: bool = False
):
    """Tell a Pydantic base model to generate its docstrings on the fly with the tech here """
    try:
        target.__doc__ = AutoPydanticDocGenerator(target, allow_failure=allow_failure)  # type: ignore
    except AutoDocError:
        if ignore_reapply:
            pass
        else:
            raise
    # Reapply by force to allow inherited models to auto doc as well
    if force_reapply:
        del target.__doc__
        auto_gen_docs_on_demand(target, allow_failure=allow_failure, ignore_reapply=ignore_reapply, force_reapply=False)


def get_base_docs(target: object):
    """Get the non-auto formatted docs, if present, otherwise just get the basic docstring"""
    if hasattr(target, AutoPydanticDocGenerator.AUTODOC_BASE_DOC_REFERENCE_ATTR):
        return getattr(target, AutoPydanticDocGenerator.AUTODOC_BASE_DOC_REFERENCE_ATTR)
    return target.__doc__



import socket
from contextlib import contextmanager

import pytest

import qcelemental
from qcelemental.util import which_import


def internet_connection():
    try:
        socket.create_connection(("www.google.com", 80))
        return True
    except OSError:
        return False


using_web = pytest.mark.skipif(internet_connection() is False, reason="Could not connect to the internet")

using_msgpack = pytest.mark.skipif(
    which_import("msgpack", return_bool=True) is False,
    reason="Not detecting module msgpack. Install package if necessary and add to envvar PYTHONPATH",
)

using_networkx = pytest.mark.skipif(
    which_import("networkx", return_bool=True) is False,
    reason="Not detecting module networkx. Install package if necessary and add to envvar PYTHONPATH",
)

using_scipy = pytest.mark.skipif(
    which_import("scipy", return_bool=True) is False,
    reason="Not detecting module scipy. Install package if necessary and add to envvar PYTHONPATH",
)

using_nglview = pytest.mark.skipif(
    which_import("nglview", return_bool=True) is False,
    reason="Not detecting module py3Dmol. Install package if necessary and add to envvar PYTHONPATH",
)

serialize_extensions = ["json", "json-ext", pytest.param("msgpack-ext", marks=using_msgpack)]


@contextmanager
def xfail_on_pubchem_busy():
    try:
        yield
    except qcelemental.ValidationError as e:
        if "HTTP Error 503: PUGREST.ServerBusy" in e.message:
            pytest.xfail("Pubchem server busy")
        else:
            raise e



"""
Datum Object Model
"""

from decimal import Decimal
from typing import Any, Dict, Optional

import numpy as np
from pydantic import BaseModel, validator


class Datum(BaseModel):
    """Facilitates the storage of quantum chemical results by labeling them with basic metadata.

    Attributes
    ----------
    label : str
        Official label for `data`, often qcvar. May contain spaces.
    units : str
        ASCII, LaTeX-like representation of units, without square brackets.
    data : float or Decimal or or :py:class:`numpy.ndarray`
        Value for `label`.
    comment : str, optional
        Additional notes.
    doi : str, optional
        Literature citation or definition DOI link.
    glossary : str, optional
        Extended description or definition.
    numeric : bool, optional
        Whether `data` is numeric. Pass `True` to disable validating `data` as float/Decimal/np.ndarray.

    """

    numeric: bool
    label: str
    units: str
    data: Any
    comment: str = ""
    doi: Optional[str] = None
    glossary: str = ""

    class Config:
        extra = "forbid"
        allow_mutation = False
        json_encoders = {np.ndarray: lambda v: v.flatten().tolist(), complex: lambda v: (v.real, v.imag)}

    def __init__(self, label, units, data, *, comment=None, doi=None, glossary=None, numeric=True):
        kwargs = {"label": label, "units": units, "data": data, "numeric": numeric}
        if comment is not None:
            kwargs["comment"] = comment
        if doi is not None:
            kwargs["doi"] = doi
        if glossary is not None:
            kwargs["glossary"] = glossary

        super().__init__(**kwargs)

    @validator("data")
    def must_be_numerical(cls, v, values, **kwargs):
        try:
            1.0 * v
        except TypeError:
            try:
                Decimal("1.0") * v
            except TypeError:
                if values["numeric"]:
                    raise ValueError(f"Datum data should be float, Decimal, or np.ndarray, not {type(v)}.")
            else:
                values["numeric"] = True
        else:
            values["numeric"] = True

        return v

    def __str__(self, label=""):
        width = 40
        text = ["-" * width, "{:^{width}}".format("Datum " + self.label, width=width)]
        if label:
            text.append("{:^{width}}".format(label, width=width))
        text.append("-" * width)
        text.append("Data:     {}".format(self.data))
        text.append("Units:    [{}]".format(self.units))
        text.append("doi:      {}".format(self.doi))
        text.append("Comment:  {}".format(self.comment))
        text.append("Glossary: {}".format(self.glossary))
        text.append("-" * width)
        return "\n".join(text)

    def dict(self, *args, **kwargs):
        return super().dict(*args, **{**kwargs, **{"exclude_unset": True}})

    def to_units(self, units=None):
        from .physical_constants import constants

        to_unit = self.units if units is None else units
        factor = constants.conversion_factor(self.units, to_unit)

        if isinstance(self.data, Decimal):
            return factor * float(self.data)
        else:
            return factor * self.data


def print_variables(qcvars: Dict[str, "Datum"]) -> str:
    """Form a printable representation of qcvariables.

    Parameters
    ----------
    qcvars : dict of Datum
        Group of Datum objects to print.

    Returns
    -------
    str
        Printable string representation of label, data, and unit in Datum-s.

    """
    text = ["\n  Variable Map:", "  ----------------------------------------------------------------------------"]

    if len(qcvars) == 0:
        text.append("  (none)")
        return "\n".join(text)

    largest_key = max(len(k) for k in qcvars) + 2  # for quotation marks
    largest_characteristic = 8
    for k, v in qcvars.items():
        try:
            exp = int(str(v.data).split("E")[1])
        except IndexError:
            pass
        else:
            largest_characteristic = max(exp, largest_characteristic)

    for k, qca in sorted(qcvars.items()):
        # if k != qca.lbl:
        #    raise ValidationError('Huh? {} != {}'.format(k, qca.label))

        if isinstance(qca.data, np.ndarray):
            data = np.array_str(qca.data, max_line_width=120, precision=8, suppress_small=True)
            data = "\n".join("        " + ln for ln in data.splitlines())
            text.append(
                """  {:{keywidth}} => {:{width}} [{}]""".format(
                    '"' + k + '"', "", qca.units, keywidth=largest_key, width=largest_characteristic + 14
                )
            )
            text.append(data)
        elif isinstance(qca.data, Decimal):
            text.append(
                """  {:{keywidth}} => {:{width}} [{}]""".format(
                    '"' + k + '"', qca.data, qca.units, keywidth=largest_key, width=largest_characteristic + 14
                )
            )
        elif not qca.numeric:
            text.append(
                """  {:{keywidth}} => {:>{width}} [{}]""".format(
                    '"' + k + '"', str(qca.data), qca.units, keywidth=largest_key, width=largest_characteristic + 14
                )
            )
        else:
            text.append(
                """  {:{keywidth}} => {:{width}.{prec}f} [{}]""".format(
                    '"' + k + '"', qca.data, qca.units, keywidth=largest_key, width=largest_characteristic + 14, prec=12
                )
            )

    text.append("")
    return "\n".join(text)



"""
Periodic table class
"""

import collections
from decimal import Decimal
from typing import Union

from .exceptions import NotAnElementError


class PeriodicTable:
    """Nuclear and mass data about chemical elements from NIST.

    Parameters
    ----------
    None

    Attributes
    ----------
    A : list of int
        Mass number, number of protons and neutrons, starting with 0 for dummies.
    Z : list of int
        Atomic number, number of protons, starting with 0 for dummies.
    E : list of str
        Element symbol from periodic table, starting with "X" for dummies. "Fe" capitalization.
    EA : list of str
        Nuclide symbol in E + A form, e.g., "Li6".
        List `EA` is a superset of `E`; that is, both "Li6" and "Li" present.
        For hydrogen, "D" and "T" also included.
    mass : list of :py:class:`decimal.Decimal`
        Atomic mass [u].
        For nuclides (e.g., "Li6"), the reported mass.
        For stable elements (e.g., "Li"), the mass of the most abundant isotope ("Li7").
        For unstable elements (e.g., "Pu"), the mass of the longest-lived isotope ("Pu244").
    name : list of str
        Element name from periodic table, starting with "Dummy". "Iron" capitalization.

    """

    def __init__(self):

        from . import data

        # Of length number of elements
        self.Z = data.nist_2011_atomic_weights["Z"]
        self.E = data.nist_2011_atomic_weights["E"]
        self.name = data.nist_2011_atomic_weights["name"]

        self._el2z = dict(zip(self.E, self.Z))
        self._z2el = collections.OrderedDict(zip(self.Z, self.E))
        self._element2el = dict(zip(self.name, self.E))
        self._el2element = dict(zip(self.E, self.name))

        # Of length number of isotopes
        self._EE = data.nist_2011_atomic_weights["_EE"]
        self.EA = data.nist_2011_atomic_weights["EA"]
        self.A = data.nist_2011_atomic_weights["A"]
        self.mass = data.nist_2011_atomic_weights["mass"]

        self._eliso2mass = dict(zip(self.EA, self.mass))
        self._eliso2el = dict(zip(self.EA, self._EE))
        self._eliso2a = dict(zip(self.EA, self.A))
        self._el2a2mass = collections.defaultdict(dict)
        for EE, m, A in zip(self._EE, self.mass, self.A):
            self._el2a2mass[EE][A] = float(m)

    def _resolve_atom_to_key(self, atom: Union[int, str], strict: bool = False) -> str:
        """Given `atom` as element name, element symbol, nuclide symbol, atomic number, or atomic number string,
        return valid `self._eliso2mass` key, regardless of case. Raises `NotAnElementError` if unidentifiable.

        """

        def resolve_eliso(atom):
            try:
                self._eliso2mass[atom.capitalize()]  # type: ignore
            except (KeyError, AttributeError):
                try:
                    E = self._z2el[int(atom)]
                except (KeyError, ValueError):
                    try:
                        E = self._element2el[atom.capitalize()]  # type: ignore
                    except (KeyError, AttributeError):
                        raise NotAnElementError(atom)
                    else:
                        return E
                else:
                    return E
            else:
                assert isinstance(atom, str)
                return atom.capitalize()

        eliso = resolve_eliso(atom)

        if strict and eliso not in self.E:
            raise NotAnElementError(eliso, strict=strict)

        return eliso

    def to_mass(self, atom: Union[int, str], *, return_decimal: bool = False) -> Union[float, "Decimal"]:
        """Get atomic mass of `atom`.

        Parameters
        ----------
        atom : int or str
            Identifier for element or nuclide, e.g., `H`, `D`, `H2`, `He`, `hE4`.
        return_decimal : bool, optional
            Whether to preserve significant figures information by returning as Decimal (`True`) or to convert to float (`False`).

        Returns
        -------
        decimal.Decimal or float
            Atomic mass [u]. See above for type.
            If `atom` is nuclide (e.g., "Li6"), the reported mass.
            If `atom` is stable element (e.g., "Li"), the mass of the most abundant isotope, "Li7".
            If `atom` is unstable element (e.g., "Pu"), the mass of the longest-lived isotope, "Pu244".

        Raises
        ------
        NotAnElementError
            If `atom` cannot be resolved into an element or nuclide.

        """
        identifier = self._resolve_atom_to_key(atom)
        mass = self._eliso2mass[identifier]

        if return_decimal:
            return Decimal(mass)
        else:
            return float(mass)

    def to_A(self, atom: Union[int, str]) -> int:
        """Get mass number of `atom`.

        Functions :py:func:`to_A` and :py:func:`to_mass_number` are aliases.

        Parameters
        ----------
        atom : int or str
            Identifier for element or nuclide, e.g., `H`, `D`, `H2`, `He`, `hE4`.

        Returns
        -------
        int
            Mass number, number of protons and neutrons.
            If `atom` is nuclide (e.g., "Li6"), the corresponding mass number, 6.
            If `atom` is stable element (e.g., "Li"), the mass number of the most abundant isotope, 7.
            If `atom` is unstable element (e.g., "Pu"), the mass number of the longest-lived isotope, 244.

        Raises
        ------
        NotAnElementError
            If `atom` cannot be resolved into an element or nuclide.

        """
        identifier = self._resolve_atom_to_key(atom)
        return self._eliso2a[identifier]

    def to_Z(self, atom: Union[int, str], strict: bool = False) -> int:
        """Get atomic number of `atom`.

        Functions :py:func:`to_Z` and :py:func:`to_atomic_number` are aliases.

        Parameters
        ----------
        atom : int or str
            Identifier for element or nuclide, e.g., `H`, `D`, `H2`, `He`, `hE4`.
        strict
            Allow only element identification in `atom`, not nuclide.

        Returns
        -------
        int
            Atomic number, number of protons.

        Raises
        ------
        NotAnElementError
            If `atom` cannot be resolved into an element or nuclide.
            If `strict=True` and `atom` resolves into nuclide, not element.

        """
        identifier = self._resolve_atom_to_key(atom, strict=strict)
        return self._el2z[self._eliso2el[identifier]]

    def to_E(self, atom: Union[int, str], strict: bool = False) -> str:
        """Get element symbol of `atom`.

        Functions :py:func:`to_E` and :py:func:`to_symbol` are aliases.

        Parameters
        ----------
        atom : Union[int, str]
            Identifier for element or nuclide, e.g., `H`, `D`, `H2`, `He`, `hE4`.
        strict
            Allow only element identification in `atom`, not nuclide.

        Returns
        -------
        str
            Element symbol, capitalized.

        Raises
        ------
        NotAnElementError
            If `atom` cannot be resolved into an element or nuclide.
            If `strict=True` and `atom` resolves into nuclide, not element.

        """
        identifier = self._resolve_atom_to_key(atom, strict=strict)
        return self._eliso2el[identifier]

    def to_element(self, atom: Union[int, str], strict: bool = False) -> str:
        """Get element name of `atom`.

        Functions :py:func:`to_element` and :py:func:`to_name` are aliases.

        Parameters
        ----------
        atom : int or str
            Identifier for element or nuclide, e.g., `H`, `D`, `H2`, `He`, `hE4`.
        strict
            Allow only element identification in `atom`, not nuclide.

        Returns
        -------
        str
            Element name, capitalized.

        Raises
        ------
        NotAnElementError
            If `atom` cannot be resolved into an element or nuclide.
            If `strict=True` and `atom` resolves into nuclide, not element.

        """
        identifier = self._resolve_atom_to_key(atom, strict=strict)
        return self._el2element[self._eliso2el[identifier]]

    to_mass_number = to_A
    to_atomic_number = to_Z
    to_symbol = to_E
    to_name = to_element

    def to_period(self, atom: Union[int, str]) -> int:
        """Get period (horizontal row in periodic table) of `atom`.

        Parameters
        ----------
        atom : int or str
            Identifier for element or nuclide, e.g., `H`, `D`, `H2`, `He`, `hE4`.

        Returns
        -------
        int
            Period between 1 (e.g., `He`) and 7 (e.g., `U238`).

        Raises
        ------
        NotAnElementError
            If `atom` cannot be resolved into an element or nuclide.

        """
        Z = self.to_Z(atom)

        if Z <= 2:
            return 1
        elif Z <= 10:
            return 2
        elif Z <= 18:
            return 3
        elif Z <= 36:
            return 4
        elif Z <= 54:
            return 5
        elif Z <= 86:
            return 6
        elif Z <= 118:
            return 7
        else:
            return 8

    def to_group(self, atom: Union[int, str]) -> Union[int, None]:
        """Get group (vertical column in periodic table) of `atom`.

        Parameters
        ----------
        atom : int or str
            Identifier for element or nuclide, e.g., `H`, `D`, `H2`, `He`, `hE4`.

        Returns
        -------
        int
            Group between 1 (e.g., `Li`) and 18 (e.g., `KR84`).
        None
            If one of the 30 Lanthanides (Z=57-71) or Actinides (Z=89-103).

        Raises
        ------
        NotAnElementError
            If `atom` cannot be resolved into an element or nuclide.

        """
        Z = self.to_Z(atom)

        # yapf: disable
        if Z in    [1, 3, 11, 19, 37, 55, 87]:
            return 1
        elif Z in     [4, 12, 20, 38, 56, 88]:
            return 2
        elif Z in            [21, 39]:
            return 3
        elif Z in            [22, 40, 72, 104]:
            return 4
        elif Z in            [23, 41, 73, 105]:
            return 5
        elif Z in            [24, 42, 74, 106]:
            return 6
        elif Z in            [25, 43, 75, 107]:
            return 7
        elif Z in            [26, 44, 76, 108]:
            return 8
        elif Z in            [27, 45, 77, 109]:
            return 9
        elif Z in            [28, 46, 78, 110]:
            return 10
        elif Z in            [29, 47, 79, 111]:
            return 11
        elif Z in            [30, 48, 80, 112]:
            return 12
        elif Z in     [5, 13, 31, 49, 81, 113]:
            return 13
        elif Z in     [6, 14, 32, 50, 82, 114]:
            return 14
        elif Z in     [7, 15, 33, 51, 83, 115]:
            return 15
        elif Z in     [8, 16, 34, 52, 84, 116]:
            return 16
        elif Z in     [9, 17, 35, 53, 85, 117]:
            return 17
        elif Z in [2, 10, 18, 36, 54, 86, 118]:
            return 18
        else:
            return None
        # yapf: enable


def run_comparison():
    """Compare the existing element information for Psi4 and Cfour (in checkup_data folder) to object. Specialized use."""

    self = PeriodicTable()

    try:
        from . import checkup_data
    except ImportError:  # pragma: no cover
        print("Info for comparison (directory checkup_data) not installed. Run from source.")

    class bcolors:
        HEADER = "\033[95m"
        OKBLUE = "\033[94m"
        OKGREEN = "\033[92m"
        WARNING = "\033[93m"
        FAIL = "\033[91m"
        ENDC = "\033[0m"
        BOLD = "\033[1m"
        UNDERLINE = "\033[4m"

    # print(bcolors.OKBLUE + '\nChecking z2element vs. Psi4 ...' + bcolors.ENDC)
    # for zz in self._z2element:
    #    if zz > 107:
    #        break
    #    assert self._z2element[zz] == checkup_data.periodictable.z2element[
    #        zz].capitalize(), 'Element {} differs from {} for Z={}'.format(
    #            z_to_element[zz], checkup_data.periodictable.z2element[zz].capitalize(), zz)

    print(bcolors.OKBLUE + "\nChecking z2el vs. Psi4 ..." + bcolors.ENDC)
    for zz in self._z2el:
        if 0 < zz < 108:
            assert (
                self._z2el[zz] == checkup_data.periodictable.z2el[zz].capitalize()
            ), "Element {} differs from {} for Z={}".format(
                self._z2el[zz], checkup_data.periodictable.z2el[zz].capitalize(), zz
            )

    print(bcolors.OKBLUE + "\nChecking el2z vs. Psi4 ..." + bcolors.ENDC)
    for el in self._el2z:
        if self._el2z[el] > 107:
            break
        if el not in ["X", "Gh"]:
            assert self._el2z[el] == checkup_data.periodictable.el2z[el.upper()], "Element {} differs from {}".format(
                self._el2z[el], checkup_data.periodictable.el2z[el.upper()]
            )

    translate = {"UUB": "Cn", "UUT": "Nh", "UUQ": "Fl", "UUP": "Mc", "UUH": "Lv", "UUS": "Ts", "UUO": "Og"}
    translate = {v: k for k, v in translate.items()}

    tol = 1.0e-5
    print(bcolors.OKBLUE + "\nChecking ({}) el2mass vs. Psi4 ...".format(tol))
    for el in self.E:
        ptel = translate.get(el, el.upper())
        if el not in ["X", "Gh"]:
            ref = self._eliso2mass[el]
            val = checkup_data.periodictable.el2mass[ptel]
            diff = abs(float(ref) - val)
            if diff > 1.0e-2:
                print(
                    bcolors.FAIL
                    + "Element {} differs by {:12.8f}: {} (this) vs {} (psi)".format(el, diff, ref, val)
                    + bcolors.ENDC
                )
            elif diff > tol:
                print("Element {} differs by {:12.8f}: {} (this) vs {} (psi)".format(el, diff, ref, val))

    tol = 1.0e-5
    print(bcolors.OKBLUE + "\nChecking ({}) el2mass vs. Cfour ...".format(tol) + bcolors.ENDC)
    for el in self.E:
        zz = self._el2z[el]
        if zz > 112:
            break
        if el not in ["X", "Gh"]:
            ref = self._eliso2mass[el]
            val = checkup_data.cfour_primary_masses[zz - 1]
            diff = abs(float(ref) - val)
            if diff > 1.0e-2:
                print(
                    bcolors.FAIL
                    + "Element {} differs by {:12.8f}: {} (this) vs {} (cfour)".format(el, diff, ref, val)
                    + bcolors.ENDC
                )
            elif diff > tol:
                print("Element {} differs by {:12.8f}: {} (this) vs {} (cfour)".format(el, diff, ref, val))

    tol = 1.0e-3
    print(bcolors.OKBLUE + "\nChecking ({}) eliso2mass vs. Psi4 ...".format(tol) + bcolors.ENDC)
    for el in self._eliso2mass:
        ptel = translate.get(el, el.upper())
        if el not in ["X", "Gh"]:
            if ptel not in checkup_data.periodictable.eliso2mass:
                print("Element {:6} missing from Psi4".format(el))
                continue
            ref = self._eliso2mass[el]
            val = checkup_data.periodictable.eliso2mass[ptel]
            diff = abs(float(ref) - val)
            if diff > 1.0e-2:
                print(
                    bcolors.FAIL
                    + "Element {:6} differs by {:12.8f}: {} (this) vs {} (psi)".format(el, diff, ref, val)
                    + bcolors.ENDC
                )
            elif diff > tol:
                print("Element {:6} differs by {:12.8f}: {} (this) vs {} (psi)".format(el, diff, ref, val))


def write_c_header(filename: str = "masses.h") -> None:
    """Write C header file defining arrays of mass and element information.

    Parameters
    ----------
    filename : str, optional
        The filename to write to.
    """
    self = PeriodicTable()

    text = [
        "#ifndef _qcelemental_masses_h_",
        "#define _qcelemental_masses_h_",
        "",
        "/* This file is autogenerated from the QCElemental python module */",
        "",
        "static const char *atomic_labels[]={",
        '"' + '","'.join(e.upper() for e in self.E) + '"',
        "};",
        "",
        "static const double an2masses[]={",
        ",".join(str(self._eliso2mass[e]) for e in self.E),
        "};",
        "",
        "static const char *mass_labels[]={",
        '"' + '","'.join(e.upper() for e in self.EA if e not in ["Gh", "X", "X0"]) + '"',
        "};",
        "",
        "static const double atomic_masses[]={",
        ",".join(str(self._eliso2mass[e]) for e in self.EA if e not in ["Gh", "X", "X0"]),
        "};",
        "",
        "#endif /* header guard */",
        "",
    ]  # yapf: disable

    with open(filename, "w") as handle:
        handle.write("\n".join(text))
    print("File written ({}). Remember to add license and clang-format it.".format(filename))


# el2mass["GH"] = 0.  # note that ghost atoms in Cfour have mass 100.
# eliso2mass["GH"] = 0.  # note that ghost atoms in Cfour have mass 100.  # encompasses el2mass
# eliso2mass["X0"] = 0.  # probably needed, just checking
# el2z["GH"] = 0

# singleton
periodictable = PeriodicTable()



import numpy as np


def random_rotation_matrix(deflection: float = 1.0, randnums=None) -> np.ndarray:
    """Generates a random 3D rotation matrix.

    Parameters
    ----------
    deflection : float, optional
        Magnitude of the rotation. For 0, no rotation; for 1, competely random
        rotation. Small deflection => small perturbation.
    randnums : array, optional
        3 random numbers in the range [0, 1]. If `None`, they will be auto-generated.

    Returns
    -------
    3 x 3 ndarray
        Rotation matrix on random distribution.

    Sources
    -------
    improved: http://demonstrations.wolfram.com/sourcecode.html?demoname=SamplingAUniformlyRandomRotation&demodisplayname=Sampling%20a%20Uniformly%20Random%20Rotation
    py code: from http://blog.lostinmyterminal.com/python/2015/05/12/random-rotation-matrix.html
    c code: from http://www.realtimerendering.com/resources/GraphicsGems/gemsiii/rand_rotation.c
    orig: James Arvo, Graphics Gems III (IBM Version) 1992, Pages 117-120, https://doi.org/10.1016/B978-0-08-050755-2.50034-8

    """
    if randnums is None:
        randnums = np.random.uniform(size=(3,))

    theta, phi, z = randnums

    # rotation about the pole (Z)
    #   from Wolfram, improved by subtr half so rotation unbiased
    theta = (theta - 1 / 2) * deflection * 2 * np.pi
    # direction of pole deflection
    phi = phi * 2 * np.pi
    # magnitude of pole deflection
    z = z * 2 * deflection

    # Compute a vector V used for distributing points over the sphere
    # via the reflection I - V Transpose(V).  This formulation of V
    # will guarantee that if x[1] and x[2] are uniformly distributed,
    # the reflected points will be uniform on the sphere.  Note that V
    # has length sqrt(2) to eliminate the 2 in the Householder matrix.
    r = np.sqrt(z)
    V = (np.sin(phi) * r, np.cos(phi) * r, np.sqrt(2.0 - z))

    st = np.sin(theta)
    ct = np.cos(theta)

    R = np.array(((ct, st, 0), (-st, ct, 0), (0, 0, 1)))
    R_z_pi = np.diag([-1.0, -1.0, 1])

    # Construct the rotation matrix  (V Transpose(V) - I) R. * R_z(pi)
    #   which is equivalent to V.S - R.
    # From Wolfram, Arno's code is missing the multiplication by R_z(pi),
    #   which is unnoticable for random rotations but causes problems
    #   for random perturbations.
    M = (np.outer(V, V) - np.eye(3)).dot(R).dot(R_z_pi)

    return M



import pytest

import qcelemental

co_dominant = (59, 27, "Co", 58.93319429, True, "")
co_dominant_mine = (59, 27, "Co", 58.93319429, True, "_mine")
co_dominant_shortmass = (59, 27, "Co", 58.933, True, "")
co60 = (60, 27, "Co", 59.93381630, True, "")
co60_mine = (60, 27, "Co", 59.93381630, True, "_mine")
co60ghost = (60, 27, "Co", 59.93381630, False, "")
co_unspecified = (-1, 27, "Co", 60.6, True, "")


@pytest.mark.parametrize(
    "inp,expected",
    [
        ({"E": "co"}, co_dominant),
        ({"Z": 27}, co_dominant),
        ({"A": 59, "Z": 27}, co_dominant),
        ({"E": "cO", "mass": 58.93319429}, co_dominant),
        ({"A": 59, "Z": 27, "E": "CO"}, co_dominant),
        ({"A": 59, "E": "cO", "mass": 58.93319429}, co_dominant),
        ({"label": "co", "verbose": 2}, co_dominant),
        ({"label": "59co"}, co_dominant),
        ({"label": "co@58.93319429"}, co_dominant),
        ({"A": 59, "Z": 27, "E": "cO", "mass": 58.93319429, "label": "co@58.93319429"}, co_dominant),
        ({"A": 59, "Z": 27, "E": "cO", "mass": 58.93319429, "label": "27@58.93319429"}, co_dominant),
        ({"label": "27"}, co_dominant),
        ({"label": "co_miNe"}, co_dominant_mine),
        ({"label": "co_mIne@58.93319429"}, co_dominant_mine),
        ({"E": "cO", "mass": 58.933}, co_dominant_shortmass),
        ({"label": "cO@58.933"}, co_dominant_shortmass),
        ({"E": "Co", "A": 60}, co60),
        ({"Z": 27, "A": 60, "real": True}, co60),
        ({"E": "Co", "A": 60}, co60),
        ({"Z": 27, "mass": 59.93381630}, co60),
        ({"A": 60, "Z": 27, "mass": 59.93381630}, co60),
        ({"label": "60Co"}, co60),
        ({"label": "27", "mass": 59.93381630}, co60),
        ({"label": "Co", "mass": 59.93381630}, co60),
        ({"A": 60, "label": "Co"}, co60),
        ({"mass": 60.6, "Z": 27}, co_unspecified),
        ({"mass": 60.6, "E": "Co"}, co_unspecified),
        ({"mass": 60.6, "label": "27"}, co_unspecified),
        ({"label": "Co@60.6"}, co_unspecified),
        ({"E": "Co", "A": 60, "real": False}, co60ghost),
        ({"A": 60, "Z": 27, "mass": 59.93381630, "real": 0}, co60ghost),
        ({"label": "@60Co"}, co60ghost),
        ({"label": "Gh(27)", "mass": 59.93381630}, co60ghost),
        ({"label": "@Co", "mass": 59.93381630}, co60ghost),
        ({"A": 60, "label": "Gh(Co)"}, co60ghost),
        ({"label": "60co_miNe"}, co60_mine),
        ({"label": "_miNe", "A": 59, "E": "Co", "speclabel": False}, co_dominant_mine),
    ],
)
def test_reconcile_nucleus(inp, expected):
    assert expected == qcelemental.molparse.reconcile_nucleus(**inp)


@pytest.mark.parametrize("inp", [{"E": "cO", "mass": 58.933, "mtol": 1.0e-4}, {"label": "27@58.933", "mtol": 1.0e-4}])
def test_reconcile_nucleus_assertionerror(inp):
    with pytest.raises(AssertionError):
        assert co_dominant_shortmass == qcelemental.molparse.reconcile_nucleus(**inp)


@pytest.mark.parametrize(
    "inp",
    [{"A": 80, "Z": 27}, {"Z": -27, "mass": 200, "nonphysical": True}, {"A": 100, "E": "He", "nonphysical": True}],
)
def test_reconcile_nucleus_notanelementerror(inp):
    with pytest.raises(qcelemental.NotAnElementError):
        qcelemental.molparse.reconcile_nucleus(**inp)


def test_reconcile_nucleus_41():
    ans = qcelemental.molparse.reconcile_nucleus(Z=27, mass=200, nonphysical=True)
    assert ans == (-1, 27, "Co", 200.0, True, "")

    ans = qcelemental.molparse.reconcile_nucleus(A=None, Z=None, E="He", mass=100, label=None, nonphysical=True)
    assert ans == (-1, 2, "He", 100.0, True, "")


@pytest.mark.parametrize(
    "inp",
    [
        {"mass": 60.6, "Z": 27, "A": 61},
        {"Z": 27, "mass": 200},
        {"Z": 27, "mass": -200, "nonphysical": True},
        {"Z": 1, "label": "he"},
        {"A": 4, "label": "3he"},
        {"label": "@U", "real": True},
        {"label": "U", "real": False},
        {"label": "_miNe", "A": 59, "E": "Co", "speclabel": True},
    ],
)
def test_reconcile_nucleus_validationerror(inp):
    with pytest.raises(qcelemental.ValidationError):
        qcelemental.molparse.reconcile_nucleus(**inp)



from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pytest
from pydantic import BaseModel, Field

import qcelemental as qcel
from qcelemental.testing import compare_recursive, compare_values

from .addons import serialize_extensions


@pytest.fixture(scope="function")
def doc_fixture():
    class Nest(BaseModel):
        """A nested model"""

        n: float = 56

    class X(BaseModel):
        """A Pydantic model made up of many, many different combinations of ways of mapping types in Pydantic"""

        x: int
        y: str = Field(...)
        n: Nest
        n2: Nest = Field(Nest(), description="A detailed description")
        z: float = 5
        z2: float = None
        z3: Optional[float]
        z4: Optional[float] = Field(5, description="Some number I just made up")
        z5: Optional[Union[float, int]]
        z6: Optional[List[int]]
        l: List[int]
        l2: List[Union[int, str]]
        t: Tuple[str, int]
        t2: Tuple[List[int]]
        t3: Tuple[Any]
        d: Dict[str, Any]
        dlu: Dict[Union[int, str], List[Union[int, str, float]]] = Field(..., description="this is complicated")
        dlu2: Dict[Any, List[Union[int, str, float]]]
        dlu3: Dict[str, Any]
        si: int = Field(..., description="A level of constraint", gt=0)
        sf: float = Field(None, description="Optional Constrained Number", le=100.3)

    yield X


@pytest.mark.parametrize("inp,expected", [(["AAAABBBCCDAABBB"], "ABCD"), (["ABBCcAD", str.lower], "ABCD")])
def test_unique_everseen(inp, expected):
    ue = qcel.util.unique_everseen(*inp)
    assert list(ue) == list(expected)


@pytest.mark.parametrize(
    "inp,expected",
    [
        (
            ({1: {"a": "A"}, 2: {"b": "B"}}, {2: {"c": "C"}, 3: {"d": "D"}}),
            {1: {"a": "A"}, 2: {"b": "B", "c": "C"}, 3: {"d": "D"}},
        ),
        (
            ({1: {"a": "A"}, 2: {"b": "B", "c": None}}, {2: {"c": "C"}, 3: {"d": "D"}}),
            {1: {"a": "A"}, 2: {"b": "B", "c": "C"}, 3: {"d": "D"}},
        ),
        (({1: [None, 1]}, {1: [2, 1], 3: {"d": "D"}}), {1: [2, 1], 3: {"d": "D"}}),
    ],
)
def test_updatewitherror(inp, expected):
    assert compare_recursive(expected, qcel.util.update_with_error(inp[0], inp[1]))


@pytest.mark.parametrize(
    "inp",
    [
        ({1: {"a": "A"}, 2: {"b": "B"}}, {1: {"a": "A"}, 2: {"b": "C"}}),
        ({1: {"a": "A"}, 2: {"b": "B"}}, {1: {"a": "A"}, 2: {"b": None}}),
        ({1: [None, 1]}, {1: [2, 2], 3: {"d": "D"}}),
    ],
)
def test_updatewitherror_error(inp):
    with pytest.raises(KeyError):
        qcel.util.update_with_error(inp[0], inp[1])


@pytest.mark.parametrize(
    "inp,expected",
    [
        ({"dicary": {"a": "A", "b": "B"}}, {"a": "A", "b": "B"}),
        ({"dicary": {"a": np.arange(2), "b": [1, 2]}}, {"a": [0, 1], "b": [1, 2]}),
        ({"dicary": {"c": {"a": np.arange(2), "b": [1, 2]}}}, {"c": {"a": [0, 1], "b": [1, 2]}}),
        ({"dicary": {"c": {"a": np.arange(2), "b": [1, 2]}}, "flat": True}, {"c": {"a": [0, 1], "b": [1, 2]}}),
        (
            {"dicary": {"c": {"a": np.arange(2), "b": [1, 2]}, "d": np.arange(6).reshape((2, 3))}},
            {"c": {"a": [0, 1], "b": [1, 2]}, "d": [[0, 1, 2], [3, 4, 5]]},
        ),
        (
            {"dicary": {"c": {"a": np.arange(2), "b": [1, 2]}, "d": np.arange(6).reshape((2, 3))}, "flat": True},
            {"c": {"a": [0, 1], "b": [1, 2]}, "d": [0, 1, 2, 3, 4, 5]},
        ),
        (
            {"dicary": {"a": np.arange(2), "e": ["mouse", np.arange(4).reshape(2, 2)]}},
            {"a": [0, 1], "e": ["mouse", [[0, 1], [2, 3]]]},
        ),
        (
            {"dicary": {"a": np.arange(2), "e": ["mouse", {"f": np.arange(4).reshape(2, 2)}]}},
            {"a": [0, 1], "e": ["mouse", {"f": [[0, 1], [2, 3]]}]},
        ),
        (
            {
                "dicary": {
                    "a": np.arange(2),
                    "e": ["mouse", [np.arange(4).reshape(2, 2), {"f": np.arange(6).reshape(2, 3), "g": [[11], [12]]}]],
                }
            },
            {"a": [0, 1], "e": ["mouse", [[[0, 1], [2, 3]], {"f": [[0, 1, 2], [3, 4, 5]], "g": [[11], [12]]}]]},
        ),
        (
            {"dicary": {"a": np.arange(2), "e": ["mouse", np.arange(4).reshape(2, 2)]}, "flat": True},
            {"a": [0, 1], "e": ["mouse", [0, 1, 2, 3]]},
        ),
        (
            {"dicary": {"a": np.arange(2), "e": ["mouse", {"f": np.arange(4).reshape(2, 2)}]}, "flat": True},
            {"a": [0, 1], "e": ["mouse", {"f": [0, 1, 2, 3]}]},
        ),
        (
            {
                "dicary": {
                    "a": np.arange(2),
                    "e": ["mouse", [np.arange(4).reshape(2, 2), {"f": np.arange(6).reshape(2, 3), "g": [[11], [12]]}]],
                },
                "flat": True,
            },
            {"a": [0, 1], "e": ["mouse", [[0, 1, 2, 3], {"f": [0, 1, 2, 3, 4, 5], "g": [[11], [12]]}]]},
        ),
    ],
)
def test_unnp(inp, expected):
    assert compare_recursive(expected, qcel.util.unnp(**inp), atol=1.0e-4)


def test_distance():
    def _test_distance(p1, p2, value):
        tmp = qcel.util.compute_distance(p1, p2)
        assert compare_values(value, float(tmp))

    _test_distance([0, 0, 0], [0, 0, 1], 1.0)
    _test_distance([0, 0, 0], [0, 0, 0], 0.0)

    tmp1 = np.random.rand(20, 3) * 4
    tmp2 = np.random.rand(20, 3) * 4
    np_dist = np.linalg.norm(tmp1 - tmp2, axis=1)
    ee_dist = qcel.util.compute_distance(tmp1, tmp2)
    assert np.allclose(np_dist, ee_dist)


def test_angle():
    def _test_angle(p1, p2, p3, value, degrees=True):
        tmp = qcel.util.compute_angle(p1, p2, p3, degrees=degrees)
        assert compare_values(value, float(tmp))

    # Check all 90 degree domains
    p1 = [5, 0, 0]
    p2 = [0, 0, 0]
    p3 = [0, 2, 0]
    p4 = [0, 0, 4]
    _test_angle(p1, p2, p3, 90)
    _test_angle(p3, p2, p1, 90)

    _test_angle(p1, p2, p4, 90)
    _test_angle(p4, p2, p1, 90)

    _test_angle(p3, p2, p4, 90)
    _test_angle(p4, p2, p3, 90)
    _test_angle(p4, p2, p3, np.pi / 2, degrees=False)

    # Zero angle
    p5 = [6, 0, 0]
    _test_angle(p1, p2, p1, 0)
    _test_angle(p1, p2, p5, 0)
    _test_angle(p1, p2, p5, 0, degrees=False)

    # Linear
    p6 = [-5, 0, 0]
    p7 = [-4, 0, 0]
    _test_angle(p1, p2, p6, 180)
    _test_angle(p1, p2, p6, 180)
    _test_angle(p1, p2, p7, np.pi, degrees=False)


def test_dihedral1():
    def _test_dihedral(p1, p2, p3, p4, value, degrees=True):
        tmp = qcel.util.compute_dihedral(p1, p2, p3, p4, degrees=degrees)
        assert compare_values(value, float(tmp), label="test_dihedral1")

    p1 = [0, 0, 0]
    p2 = [0, 2, 0]
    p3 = [2, 2, 0]
    p4 = [2, 0, 0]
    p5 = [2, 4, 0]

    # Cis check
    _test_dihedral(p1, p2, p3, p4, 0)

    # Trans check
    _test_dihedral(p1, p2, p3, p5, 180)

    # 90 phase checks
    p6 = [2, 2, -2]
    p7 = [2, 2, 2]
    _test_dihedral(p1, p2, p3, p6, 90)
    _test_dihedral(p1, p2, p3, p7, -90)

    p8 = [0, 4, 0]
    _test_dihedral(p8, p2, p3, p6, -90)
    _test_dihedral(p8, p2, p3, p7, 90)

    # Linear checks
    _test_dihedral(p1, p2, p3, [3, 2, 0], 0)
    _test_dihedral(p1, p2, p3, [3, 2 + 1.0e-14, 0], 180)
    _test_dihedral(p1, p2, p3, [3, 2 - 1.0e-14, 0], 0)
    _test_dihedral(p1, p2, p3, [3, 2, 0 + 1.0e-14], -90)
    _test_dihedral(p1, p2, p3, [3, 2, 0 - 1.0e-14], 90)


def test_dihedral2():
    # FROM: https://stackoverflow.com/questions/20305272/

    def _test_dihedral(p1, p2, p3, p4, value, degrees=True):
        tmp = qcel.util.compute_dihedral(p1, p2, p3, p4, degrees=degrees)
        assert compare_values(value, float(tmp), label="test_dihedral1")

    p0 = [24.969, 13.428, 30.692]
    p1 = [24.044, 12.661, 29.808]
    p2 = [22.785, 13.482, 29.543]
    p3 = [21.951, 13.670, 30.431]
    p4 = [23.672, 11.328, 30.466]
    p5 = [22.881, 10.326, 29.620]
    p6 = [23.691, 9.935, 28.389]
    p7 = [22.557, 9.096, 30.459]

    _test_dihedral(p0, p1, p2, p3, -71.21515114671394)
    _test_dihedral(p3, p2, p1, p0, -71.21515114671394)

    _test_dihedral(p0, p1, p4, p5, -171.94319947953642)
    _test_dihedral(p5, p4, p1, p0, -171.94319947953642)

    _test_dihedral(p1, p4, p5, p6, 60.82226735264638)
    _test_dihedral(p6, p5, p4, p1, 60.82226735264638)

    _test_dihedral(p1, p4, p5, p7, -177.63641151521261)
    _test_dihedral(p7, p5, p4, p1, -177.63641151521261)


def test_auto_gen_doc(doc_fixture):
    assert "this is complicated" not in doc_fixture.__doc__
    qcel.util.auto_gen_docs_on_demand(doc_fixture, allow_failure=False, ignore_reapply=False)
    assert "this is complicated" in doc_fixture.__doc__
    assert "z3 : float, Optional" in doc_fixture.__doc__
    # Check that docstring does not get duplicated for some reason
    assert doc_fixture.__doc__.count("z3 : float, Optional") == 1


def test_auto_gen_doc_exiting(doc_fixture):
    doc_fixture.__doc__ = "Parameters\n"
    qcel.util.auto_gen_docs_on_demand(doc_fixture, allow_failure=False, ignore_reapply=False)
    assert "this is complicated" not in doc_fixture.__doc__


def test_auto_gen_doc_reapply_failure(doc_fixture):
    qcel.util.auto_gen_docs_on_demand(doc_fixture, allow_failure=False, ignore_reapply=False)
    with pytest.raises(ValueError):
        # Allow true here because we are testing application, not errors in the doc generation itself
        qcel.util.auto_gen_docs_on_demand(doc_fixture, allow_failure=True, ignore_reapply=False)


def test_auto_gen_doc_delete(doc_fixture):
    qcel.util.auto_gen_docs_on_demand(doc_fixture, allow_failure=False, ignore_reapply=False)
    assert "this is complicated" in doc_fixture.__doc__
    assert "A Pydantic model" in doc_fixture.__doc__
    del doc_fixture.__doc__
    assert "this is complicated" not in doc_fixture.__doc__
    assert "A Pydantic model" in doc_fixture.__doc__


@pytest.mark.parametrize(
    "obj",
    [
        5,
        1.11111,
        "hello",
        "\u0394",
        np.random.rand(4),
        {"a": 5},
        {"a": 1.111111111111},
        {"a": "hello"},
        {"a": np.random.rand(4), "b": np.array(5.1111111), "c": np.array("hello")},
        ["12345"],
        ["hello", "world"],
        [5, 123.234, "abcdé", "\u0394", "\U00000394"],
        [5, "B63", np.random.rand(4)],
        ["abcdé", {"a": np.random.rand(2), "b": np.random.rand(5)}],
        [np.array(3), np.arange(3, dtype=np.uint16), {"b": np.array(["a", "b"])}],
    ],
)
@pytest.mark.parametrize("encoding", serialize_extensions)
def test_serialization(obj, encoding):
    new_obj = qcel.util.deserialize(qcel.util.serialize(obj, encoding=encoding), encoding=encoding)
    assert compare_recursive(obj, new_obj)



"""
Tests the imports and exports of the Molecule object.
"""

import numpy as np
import pytest

import qcelemental as qcel
from qcelemental.exceptions import NotAnElementError
from qcelemental.models import Molecule
from qcelemental.testing import compare, compare_values

from .addons import serialize_extensions, using_msgpack, using_nglview

water_molecule = Molecule.from_data(
    """
    0 1
    O  -1.551007  -0.114520   0.000000
    H  -1.934259   0.762503   0.000000
    H  -0.599677   0.040712   0.000000
    """
)

water_dimer_minima = Molecule.from_data(
    """
    0 1
    O  -1.551007  -0.114520   0.000000
    H  -1.934259   0.762503   0.000000
    H  -0.599677   0.040712   0.000000
    --
    O   1.350625   0.111469   0.000000
    H   1.680398  -0.373741  -0.758561
    H   1.680398  -0.373741   0.758561
    """,
    dtype="psi4",
    orient=True,
)


def test_molecule_data_constructor_numpy():
    water_psi = water_dimer_minima.copy()
    ele = np.array(water_psi.atomic_numbers).reshape(-1, 1)
    npwater = np.hstack((ele, water_psi.geometry * qcel.constants.conversion_factor("Bohr", "angstrom")))

    water_from_np = Molecule.from_data(npwater, name="water dimer", dtype="numpy", frags=[3])
    assert water_psi == water_from_np

    water_from_np = Molecule.from_data(npwater, name="water dimer", frags=[3])
    assert water_psi == water_from_np
    assert water_psi.get_molecular_formula() == "H4O2"
    assert water_psi.get_molecular_formula(order="alphabetical") == "H4O2"
    assert water_psi.get_molecular_formula(order="hill") == "H4O2"


def test_molecule_data_constructor_dict():
    water_psi = water_dimer_minima.copy()

    # Check the JSON construct/deconstruct
    water_from_json = Molecule.from_data(water_psi.dict())
    assert water_psi == water_from_json

    water_from_json = Molecule.from_data(water_psi.json(), "json")
    assert water_psi == water_from_json
    assert water_psi == Molecule.from_data(water_psi.to_string("psi4"), dtype="psi4")

    assert water_psi.get_hash() == "3c4b98f515d64d1adc1648fe1fe1d6789e978d34"  # copied from schema_version=1
    assert water_psi.schema_version == 2
    assert water_psi.schema_name == "qcschema_molecule"


def test_molecule_data_constructor_error():
    with pytest.raises(TypeError):
        Molecule.from_data([])

    with pytest.raises(KeyError):
        Molecule.from_data({}, dtype="bad")


def test_hash_canary():
    water_dimer_minima = Molecule.from_data(
        """
    0 1
    O  -1.551007  -0.114520   0.000000
    H  -1.934259   0.762503   0.000000
    H  -0.599677   0.040712   0.000000
    --
    O   1.350625   0.111469   0.000000
    H   1.680398  -0.373741  -0.758561
    H   1.680398  -0.373741   0.758561
    """,
        dtype="psi4",
    )
    assert water_dimer_minima.get_hash() == "42f3ac52af52cf2105c252031334a2ad92aa911c"

    # Check orientation
    mol = water_dimer_minima.orient_molecule()
    assert mol.get_hash() == "632490a0601500bfc677e9277275f82fbc45affe"

    frag_0 = mol.get_fragment(0, orient=True)
    frag_1 = mol.get_fragment(1, orient=True)
    assert frag_0.get_hash() == "d0b499739f763e8d3a5556b4ddaeded6a148e4d5"
    assert frag_1.get_hash() == "bdc1f75bd1b7b999ff24783d7c1673452b91beb9"


def test_molecule_np_constructors():
    """
    Neon tetramer fun
    """
    ### Neon Tetramer
    neon_from_psi = Molecule.from_data(
        """
        Ne 0.000000 0.000000 0.000000
        --
        Ne 3.100000 0.000000 0.000000
        --
        Ne 0.000000 3.200000 0.000000
        --
        Ne 0.000000 0.000000 3.300000
        units bohr""",
        dtype="psi4",
    )
    ele = np.array([10, 10, 10, 10]).reshape(-1, 1)
    npneon = np.hstack((ele, neon_from_psi.geometry))
    neon_from_np = Molecule.from_data(npneon, name="neon tetramer", dtype="numpy", frags=[1, 2, 3], units="bohr")

    assert neon_from_psi == neon_from_np

    # Check the JSON construct/deconstruct
    neon_from_json = Molecule.from_data(neon_from_psi.json(), dtype="json")
    assert neon_from_psi == neon_from_json
    assert neon_from_json.get_molecular_formula() == "Ne4"


def test_molecule_compare():
    water_molecule2 = water_molecule.copy()
    assert water_molecule2 == water_molecule

    water_molecule3 = water_molecule.copy(update={"geometry": (water_molecule.geometry + np.array([0.1, 0, 0]))})
    assert water_molecule != water_molecule3


def test_water_minima_data():
    # Give it a name
    mol_dict = water_dimer_minima.dict()
    mol_dict["name"] = "water dimer"
    mol = Molecule(orient=True, **mol_dict)

    assert len(mol.pretty_print()) == 661
    assert len(mol.to_string("psi4")) == 479

    assert sum(x == y for x, y in zip(mol.symbols, ["O", "H", "H", "O", "H", "H"])) == mol.geometry.shape[0]
    assert mol.name == "water dimer"
    assert mol.molecular_charge == 0
    assert mol.molecular_multiplicity == 1
    assert np.sum(mol.real) == mol.geometry.shape[0]
    assert np.allclose(mol.fragments, [[0, 1, 2], [3, 4, 5]])
    assert np.allclose(mol.fragment_charges, [0, 0])
    assert np.allclose(mol.fragment_multiplicities, [1, 1])
    assert hasattr(mol, "provenance")
    assert np.allclose(
        mol.geometry,
        [
            [2.81211080, 0.1255717, 0.0],
            [3.48216664, -1.55439981, 0.0],
            [1.00578203, -0.1092573, 0.0],
            [-2.6821528, -0.12325075, 0.0],
            [-3.27523824, 0.81341093, 1.43347255],
            [-3.27523824, 0.81341093, -1.43347255],
        ],
    )
    assert mol.get_hash() == "3c4b98f515d64d1adc1648fe1fe1d6789e978d34"


def test_water_minima_fragment():

    mol = water_dimer_minima.copy()
    frag_0 = mol.get_fragment(0, orient=True)
    frag_1 = mol.get_fragment(1, orient=True)
    assert frag_0.get_hash() == "5f31757232a9a594c46073082534ca8a6806d367"
    assert frag_1.get_hash() == "bdc1f75bd1b7b999ff24783d7c1673452b91beb9"

    frag_0_1 = mol.get_fragment(0, 1)
    frag_1_0 = mol.get_fragment(1, 0)

    assert np.array_equal(mol.symbols[:3], frag_0.symbols)
    assert np.allclose(mol.masses[:3], frag_0.masses)

    assert np.array_equal(mol.symbols, frag_0_1.symbols)
    assert np.allclose(mol.geometry, frag_0_1.geometry)

    assert np.array_equal(np.hstack((mol.symbols[3:], mol.symbols[:3])), frag_1_0.symbols)
    assert np.allclose(np.hstack((mol.masses[3:], mol.masses[:3])), frag_1_0.masses)


def test_pretty_print():

    mol = water_dimer_minima.copy()
    assert isinstance(mol.pretty_print(), str)


def test_to_string():

    mol = water_dimer_minima.copy()
    assert isinstance(mol.to_string("psi4"), str)


@pytest.mark.parametrize(
    "dtype, filext",
    [("json", "json"), ("xyz", "xyz"), ("numpy", "npy"), pytest.param("msgpack", "msgpack", marks=using_msgpack)],
)
def test_to_from_file_simple(tmp_path, dtype, filext):

    benchmol = Molecule.from_data(
        """
    O 0 0 0
    H 0 1.5 0
    H 0 0 1.5
    """
    )

    p = tmp_path / ("water." + filext)
    benchmol.to_file(p)

    mol = Molecule.from_file(p)

    assert mol == benchmol


@pytest.mark.parametrize("dtype", ["json", "psi4"])
def test_to_from_file_complex(tmp_path, dtype):

    p = tmp_path / ("water." + dtype)
    water_dimer_minima.to_file(p)

    mol = Molecule.from_file(p)
    assert mol == water_dimer_minima


@pytest.mark.parametrize(
    "dtype, filext", [("json", "json"), ("xyz+", "xyz"), pytest.param("msgpack", "msgpack", marks=using_msgpack)]
)
def test_to_from_file_charge_spin(tmp_path, dtype, filext):

    benchmol = Molecule.from_data(
        """
    1 2
    O 0 0 0
    H 0 1.5 0
    H 0 0 1.5
    """
    )

    p = tmp_path / ("water." + filext)
    benchmol.to_file(p, dtype=dtype)

    mol = Molecule.from_file(p, dtype=dtype)

    assert mol.molecular_charge == 1
    assert mol.molecular_multiplicity == 2
    assert mol.fragment_charges[0] == 1
    assert mol.fragment_multiplicities[0] == 2
    assert mol == benchmol


def test_from_data_kwargs():
    mol = Molecule.from_data(
        """
        O 0 0 0
        H 0 1.5 0
        H 0 0 1.5
        """,
        molecular_charge=1,
        molecular_multiplicity=2,
        fragment_charges=[1],
        fragment_multiplicities=[2],
    )
    assert mol.molecular_charge == 1
    assert mol.molecular_multiplicity == 2
    assert mol.fragment_charges[0] == 1
    assert mol.fragment_multiplicities[0] == 2

    mol = Molecule.from_data(
        """
            O 0 0 0
            H 0 1.5 0
            H 0 0 1.5
            """,
        molecular_charge=1,
        molecular_multiplicity=2,
    )
    assert mol.molecular_charge == 1
    assert mol.molecular_multiplicity == 2
    assert mol.fragment_charges[0] == 1
    assert mol.fragment_multiplicities[0] == 2

    with pytest.raises(qcel.ValidationError) as e:
        mol = Molecule.from_data(
            """
            O 0 0 0
            H 0 1.5 0
            H 0 0 1.5
            """,
            molecular_charge=1,
            molecular_multiplicity=2,
            fragment_charges=[2],
        )
    assert "Inconsistent or unspecified chg/mult" in str(e.value)


def test_water_orient():
    # These are identical molecules, should find the correct results
    mol = Molecule.from_data(
        """
        O  -1.551007  -0.114520   0.000000
        H  -1.934259   0.762503   0.000000
        H  -0.599677   0.040712   0.000000
        --
        O  -0.114520  -1.551007  10.000000
        H   0.762503  -1.934259  10.000000
        H   0.040712  -0.599677  10.000000
        """
    )

    frag_0 = mol.get_fragment(0, orient=True)
    frag_1 = mol.get_fragment(1, orient=True)

    # Make sure the fragments match
    assert frag_0.get_hash() == frag_1.get_hash()

    # Make sure the complexes match
    frag_0_1 = mol.get_fragment(0, 1, orient=True, group_fragments=True)
    frag_1_0 = mol.get_fragment(1, 0, orient=True, group_fragments=True)
    assert frag_0_1.get_hash() == frag_1_0.get_hash()

    # Fragments not reordered, should be different molecules.
    frag_0_1 = mol.get_fragment(0, 1, orient=True, group_fragments=False)
    frag_1_0 = mol.get_fragment(1, 0, orient=True, group_fragments=False)
    assert frag_0_1.get_hash() != frag_1_0.get_hash()

    # These are identical molecules, but should be different with ghost
    mol = Molecule.from_data(
        """
        O  -1.551007  -0.114520   0.000000
        H  -1.934259   0.762503   0.000000
        H  -0.599677   0.040712   0.000000
        --
        O  -11.551007  -0.114520   0.000000
        H  -11.934259   0.762503   0.000000
        H  -10.599677   0.040712   0.000000
        """,
        dtype="psi4",
        orient=True,
    )

    frag_0 = mol.get_fragment(0, orient=True)
    frag_1 = mol.get_fragment(1, orient=True)

    # Make sure the fragments match
    assert frag_0.molecular_multiplicity == 1
    assert frag_0.get_hash() == frag_1.get_hash()

    # Make sure the complexes match
    frag_0_1 = mol.get_fragment(0, 1, orient=True)
    frag_1_0 = mol.get_fragment(1, 0, orient=True)

    # Ghost fragments should prevent overlap
    assert frag_0_1.molecular_multiplicity == 1
    assert frag_0_1.get_hash() != frag_1_0.get_hash()


def test_molecule_errors_extra():
    data = water_dimer_minima.dict(exclude_unset=True)
    data["whatever"] = 5
    with pytest.raises(Exception):
        Molecule(**data, validate=False)


def test_molecule_errors_connectivity():
    data = water_molecule.dict()
    data["connectivity"] = [(-1, 5, 5)]
    with pytest.raises(Exception):
        Molecule(**data)


def test_molecule_errors_shape():
    data = water_molecule.dict()
    data["geometry"] = list(range(8))
    with pytest.raises(Exception):
        Molecule(**data)


def test_molecule_json_serialization():
    assert isinstance(water_dimer_minima.json(), str)

    assert isinstance(water_dimer_minima.dict(encoding="json")["geometry"], list)

    assert water_dimer_minima == Molecule.from_data(water_dimer_minima.json(), dtype="json")


@pytest.mark.parametrize("encoding", serialize_extensions)
def test_molecule_serialization(encoding):
    blob = water_dimer_minima.serialize(encoding)
    assert water_dimer_minima == Molecule.parse_raw(blob, encoding=encoding)


def test_charged_fragment():
    mol = Molecule(
        symbols=["Li", "Li"],
        geometry=[0, 0, 0, 0, 0, 5],
        fragment_charges=[0.0, 0.0],
        fragment_multiplicities=[2, 2],
        fragments=[[0], [1]],
    )
    assert mol.molecular_multiplicity == 3
    assert mol.molecular_charge == 0
    f1 = mol.get_fragment(0)
    assert f1.molecular_multiplicity == 2
    assert f1.fragment_multiplicities == [2]
    assert pytest.approx(f1.molecular_charge) == 0
    assert pytest.approx(f1.fragment_charges) == [0]


# someday, when we can have noncontig frags
#    mol = Molecule(**{
#        'fragments': [[1, 5, 6], [0], [2, 3, 4]],
#        'symbols': ["he", "o", "o", "h", "h", "h", "h"],
#        # same geom as test_water_orient but with He at origin
#        'geometry': [
#        0.0, 0.0, 0.0,
#        -1.551007,  -0.114520,   0.000000,
#        -0.114520,  -1.551007,  10.000000,
#         0.762503,  -1.934259,  10.000000,
#         0.040712,  -0.599677,  10.000000,
#        -1.934259,   0.762503,   0.000000,
#        -0.599677,   0.040712,   0.000000],
#    })


@pytest.mark.parametrize("group_fragments, orient", [(True, True), (False, False)])  # original  # Psi4-like
def test_get_fragment(group_fragments, orient):
    mol = Molecule(
        **{
            "fragments": [[0], [1, 2, 3], [4, 5, 6]],
            "symbols": ["he", "o", "h", "h", "o", "h", "h"],
            # same geom as test_water_orient but with He at origin
            "geometry": np.array(
                [
                    [0.0, 0.0, 0.0],
                    [-1.551007, -0.114520, 0.000000],
                    [-1.934259, 0.762503, 0.000000],
                    [-0.599677, 0.040712, 0.000000],
                    [-0.114520, -1.551007, 10.000000],
                    [0.762503, -1.934259, 10.000000],
                    [0.040712, -0.599677, 10.000000],
                ]
            )
            / qcel.constants.bohr2angstroms,
        }
    )

    assert mol.nelectrons() == 22
    assert compare_values(32.25894779318589, mol.nuclear_repulsion_energy(), atol=1.0e-5)
    assert mol.symbols[0] == "He"

    monomers_nelectrons = [2, 10, 10]
    monomers_nre = [0.0, 9.163830150548483, 9.163830150548483]
    monomers = [mol.get_fragment(ifr, group_fragments=group_fragments, orient=orient) for ifr in range(3)]
    for fr in range(3):
        assert monomers[fr].nelectrons() == monomers_nelectrons[fr]
        assert compare_values(monomers[fr].nuclear_repulsion_energy(), monomers_nre[fr], "monomer nre", atol=1.0e-5)

    idimers = [(0, 1), (0, 2), (1, 2), (1, 0), (2, 0), (2, 1)]
    dimers_nelectrons = [12, 12, 20, 12, 12, 20]
    dimers_nre = [16.8777971, 10.2097206, 23.4990904, 16.8777971, 10.2097206, 23.4990904]
    dimers = [mol.get_fragment(rl, group_fragments=group_fragments, orient=orient) for rl in idimers]
    for ifr in range(len(idimers)):
        # print('dd', ifr, idimers[ifr], dimers[ifr].nuclear_repulsion_energy(), dimers[ifr].get_hash())
        assert dimers[ifr].nelectrons() == dimers_nelectrons[ifr], "dimer nelec"
        assert compare_values(dimers[ifr].nuclear_repulsion_energy(), dimers_nre[ifr], "dimer nre", atol=1.0e-5)
    if group_fragments and orient:
        assert dimers[0].get_hash() != dimers[3].get_hash()  # atoms out of order
        assert dimers[1].get_hash() != dimers[4].get_hash()  # atoms out of order
        assert dimers[2].get_hash() == dimers[5].get_hash()
    elif not group_fragments and not orient:
        assert dimers[0].get_hash() == dimers[3].get_hash()
        assert dimers[1].get_hash() == dimers[4].get_hash()
        assert dimers[2].get_hash() == dimers[5].get_hash()
    else:
        assert 0  # lgtm: [py/unreachable-statement]

    ghdimers_nelectrons = [2, 2, 10, 10, 10, 10]
    ghdimers_nre = [0.0, 0.0, 9.163830150548483, 9.163830150548483, 9.163830150548483, 9.163830150548483]
    ghdimers = [mol.get_fragment(rl, gh, group_fragments=group_fragments, orient=orient) for rl, gh in idimers]
    for ifr in range(len(idimers)):
        # print('gh', ifr, idimers[ifr], ghdimers[ifr].nuclear_repulsion_energy(), ghdimers[ifr].get_hash())
        assert ghdimers[ifr].nelectrons() == ghdimers_nelectrons[ifr], "gh dimer nelec"
        assert compare_values(ghdimers[ifr].nuclear_repulsion_energy(), ghdimers_nre[ifr], "gh dimer nre", atol=1.0e-5)

    if group_fragments and orient:
        assert ghdimers[0].get_hash() != ghdimers[3].get_hash()  # diff atoms ghosted
        assert ghdimers[1].get_hash() != ghdimers[4].get_hash()  # diff atoms ghosted
        assert ghdimers[2].get_hash() == ghdimers[5].get_hash()
    elif not group_fragments and not orient:
        assert ghdimers[0].get_hash() != ghdimers[3].get_hash()  # diff atoms ghosted
        assert ghdimers[1].get_hash() != ghdimers[4].get_hash()  # diff atoms ghosted
        assert ghdimers[2].get_hash() != ghdimers[5].get_hash()  # real pattern different
        assert not np.allclose(ghdimers[2].real, ghdimers[5].real)
    else:
        assert 0  # lgtm: [py/unreachable-statement]


def test_molecule_repeated_hashing():

    mol = Molecule(
        **{
            "symbols": ["H", "O", "O", "H"],
            "geometry": [
                [1.7317, 1.2909, 1.037100000000001],
                [1.3156, -0.0074, -0.2807],
                [-1.3143, 0.0084, -0.2741],
                [-1.7241, -1.3079, 1.0277],
            ],
        }
    )

    h1 = mol.get_hash()
    assert mol.get_molecular_formula() == "H2O2"

    mol2 = Molecule(orient=False, **mol.dict())
    assert h1 == mol2.get_hash()

    mol3 = Molecule(orient=False, **mol2.dict())
    assert h1 == mol3.get_hash()


@pytest.mark.parametrize(
    "measure,result",
    [
        ([0, 1], 1.8086677572537304),
        ([0, 1, 2], 37.98890673587713),
        ([0, 1, 2, 3], 180.0),
        ([[0, 1, 2, 3]], [180.0]),
        ([[1, 3], [3, 1], [1, 2, 3]], [6.3282716, 6.3282716, 149.51606694803903]),
    ],
)
def test_measurements(measure, result):

    Molecule(
        **{
            "symbols": ["H", "O", "O", "H"],
            "geometry": [
                [1.7317, 1.2909, 1.0371],
                [1.3156, -0.0074, -0.2807],
                [-1.3143, 0.0084, -0.2741],
                [-1.7241, -1.3079, 1.0277],
            ],
        }
    )

    assert pytest.approx(water_dimer_minima.measure(measure)) == result


@pytest.mark.parametrize(
    "f1c,f1m,f2c,f2m,tc,tm",
    [
        (0, 2, 0, 2, 0.0, 3),
        (0, 4, 0, 2, 0.0, 5),
        (1, 1, 1, 1, 2.0, 1),
        (1, 1, 0, 2, 1.0, 2),
        (0, 2, 1, 1, 1.0, 2),
        (-1, 1, 1, 1, 0.0, 1),
    ],
)
def test_fragment_charge_configurations(f1c, f1m, f2c, f2m, tc, tm):

    mol = Molecule.from_data(
        """
    {f1c} {f1m}
    Li 0 0 0
    --
    {f2c} {f2m}
    Li 0 0 5
    """.format(
            f1c=f1c, f1m=f1m, f2c=f2c, f2m=f2m
        )
    )

    assert pytest.approx(mol.molecular_charge) == tc
    assert mol.molecular_multiplicity == tm

    # Test fragment1
    assert pytest.approx(mol.get_fragment(0).molecular_charge) == f1c
    assert mol.get_fragment(0).molecular_multiplicity == f1m

    assert pytest.approx(mol.get_fragment(0, 1).molecular_charge) == f1c
    assert mol.get_fragment(0, 1).molecular_multiplicity == f1m

    # Test fragment2
    assert pytest.approx(mol.get_fragment(1).molecular_charge) == f2c
    assert mol.get_fragment(1).molecular_multiplicity == f2m

    assert pytest.approx(mol.get_fragment([1], 0).molecular_charge) == f2c
    assert mol.get_fragment(1, [0]).molecular_multiplicity == f2m


def test_nuclearrepulsionenergy_nelectrons():

    mol = Molecule.from_data(
        """
    0 1
    --
    O          0.75119       -0.61395        0.00271
    H          1.70471       -0.34686        0.00009
    --
    1 1
    N         -2.77793        0.00179       -0.00054
    H         -2.10136        0.51768        0.60424
    H         -3.45559       -0.51904        0.60067
    H         -2.26004       -0.67356       -0.60592
    H         -3.29652        0.68076       -0.60124
    units ang
    """
    )

    assert compare_values(34.60370459, mol.nuclear_repulsion_energy(), "D", atol=1.0e-5)
    assert compare_values(4.275210518, mol.nuclear_repulsion_energy(ifr=0), "M1", atol=1.0e-5)
    assert compare_values(16.04859029, mol.nuclear_repulsion_energy(ifr=1), "M2", atol=1.0e-5)

    assert compare(20, mol.nelectrons(), "D")
    assert compare(10, mol.nelectrons(ifr=0), "M1")
    assert compare(10, mol.nelectrons(ifr=1), "M2")

    mol = mol.get_fragment([1], 0, group_fragments=False)
    # Notice the 0th/1st fragments change if default group_fragments=True.
    ifr0 = 0
    ifr1 = 1
    assert compare_values(16.04859029, mol.nuclear_repulsion_energy(), "D", atol=1.0e-5)
    assert compare_values(0.0, mol.nuclear_repulsion_energy(ifr=ifr0), "M1", atol=1.0e-5)
    assert compare_values(16.04859029, mol.nuclear_repulsion_energy(ifr=ifr1), "M2", atol=1.0e-5)

    assert compare(10, mol.nelectrons(), "D")
    assert compare(0, mol.nelectrons(ifr=ifr0), "M1")
    assert compare(10, mol.nelectrons(ifr=ifr1), "M2")


@using_nglview
def test_show():

    water_dimer_minima.show()


def test_molecule_connectivity():
    data = {"geometry": np.random.rand(5, 3), "symbols": ["he"] * 5, "validate": False}
    mol = Molecule(**data, connectivity=None)

    connectivity = [[n, n + 1, 1] for n in range(4)]
    mol = Molecule(**data, connectivity=connectivity)

    connectivity[0][0] = -1
    with pytest.raises(ValueError):
        mol = Molecule(**data, connectivity=connectivity)


def test_orient_nomasses():
    """
    Masses must be auto generated on the fly
    """

    mol = Molecule(symbols=["He", "He"], geometry=[0, 0, -2, 0, 0, 2], orient=True, validated=True)

    assert mol.__dict__["masses_"] is None
    assert compare_values([[2, 0, 0], [-2, 0, 0]], mol.geometry)


@pytest.mark.parametrize(
    "mol_string, extra_keys",
    [
        ("He 0 0 0", None),
        ("He 0 0 0\n--\nHe 0 0 5", {"fragments", "fragment_charges", "fragment_multiplicities"}),
        ("He 0 0 0\n--\n@He 0 0 5", {"fragments", "fragment_charges", "fragment_multiplicities", "real"}),
        ("He4 0 0 0", {"atom_labels"}),
        ("He@3.14 0 0 0", {"masses", "mass_numbers"}),
    ],
)
def test_sparse_molecule_fields(mol_string, extra_keys):

    expected_keys = {
        "schema_name",
        "schema_version",
        "validated",
        "symbols",
        "geometry",
        "name",
        "molecular_charge",
        "molecular_multiplicity",
        "fix_com",
        "fix_orientation",
        "provenance",
    }
    mol = Molecule.from_data(mol_string)

    if extra_keys is not None:
        expected_keys |= extra_keys

    diff_keys = mol.dict().keys() ^ expected_keys
    assert len(diff_keys) == 0, f"Diff Keys {diff_keys}"


def test_sparse_molecule_connectivity():
    """
    A bit of a weird test, but because we set connectivity it should carry through.
    """
    mol = Molecule(symbols=["He", "He"], geometry=[0, 0, -2, 0, 0, 2], connectivity=None)
    assert "connectivity" in mol.dict()
    assert mol.dict()["connectivity"] is None

    mol = Molecule(symbols=["He", "He"], geometry=[0, 0, -2, 0, 0, 2])
    assert "connectivity" not in mol.dict()


def test_bad_isotope_spec():
    with pytest.raises(NotAnElementError) as e:
        qcel.models.Molecule(symbols=["He3"], geometry=[0, 0, 0])


def test_good_isotope_spec():
    assert compare_values(
        [3.01602932], qcel.models.Molecule(symbols=["He"], mass_numbers=[3], geometry=[0, 0, 0]).masses, "nonstd mass"
    )


def test_nonphysical_spec():
    mol = qcel.models.Molecule(symbols=["He"], masses=[100], geometry=[0, 0, 0], nonphysical=True)
    assert compare_values([100.0], mol.masses, "nonphysical mass")

    print(mol.to_string(dtype="psi4"))



"""
Contains covalent radii
"""

import collections
from decimal import Decimal
from typing import Dict, Union

from .datum import Datum, print_variables
from .exceptions import DataUnavailableError
from .periodic_table import periodictable


class CovalentRadii:
    """Covalent radii sets.

    Parameters
    ----------
    context : {'ALVAREZ2008'}
        Origin of loaded data.

    Attributes
    ----------
    cr : dict of Datum
        Each covalent radius is an entry in `cr`, where key is the
        "Fe"-cased element symbol if generic or symbol-prefixed label
        if specialized within element. The value is a Datum object with
        `lbl` the same as key, `units`, `data` value as Decimal object,
        and any uncertainty in the `comment` field.
    doi : str
        The DOI of the current context.
    name : str
        The name of the context ('ALVAREZ2008')
    native_units : str
        The units the original data was provided in.
    year : int
        The year the context was created.

    """

    def __init__(self, context: str = "ALVAREZ2008"):
        self.cr: Dict[str, Datum] = collections.OrderedDict()

        from .data import alvarez_2008_covalent_radii

        if context == "ALVAREZ2008":
            self.doi = alvarez_2008_covalent_radii["doi"]
            self.native_units = alvarez_2008_covalent_radii["units"]

            # TypedDict wont be in until 3.8, have to ignore heterogeneous dicts for now
            for cr in alvarez_2008_covalent_radii["covalent_radii"]:  # type: ignore
                self.cr[cr[0]] = Datum(cr[0], self.native_units, Decimal(cr[1]), comment=cr[2], doi=self.doi)
        else:
            raise KeyError(f"Context set as '{context}', " + "only contexts {'ALVAREZ2008', } are currently supported")

        self.name = context
        self.year = int(alvarez_2008_covalent_radii["date"][:4])  # type: ignore

        # Extra relationships
        aliases = [
            ("C", "angstrom", self.cr["C_sp3"].data, "Largest (sp3) chosen for generic atom"),
            ("Mn", "angstrom", self.cr["Mn_highspin"].data, "Larger (high-spin) chosen for generic atom"),
            ("Fe", "angstrom", self.cr["Fe_highspin"].data, "Larger (high-spin) chosen for generic atom"),
            ("Co", "angstrom", self.cr["Co_highspin"].data, "Larger (high-spin) chosen for generic atom"),
        ]  # yapf: disable

        # add alternate names to help QC programs
        for alias in aliases:
            ident, units, value, comment = alias
            self.cr[ident.capitalize()] = Datum(ident, units, value, comment=comment)

    def __str__(self) -> str:
        return "CovalentRadii(context='{}')".format(self.name)

    def get(
        self, atom: Union[int, str], *, return_tuple: bool = False, units: str = "bohr", missing: float = None
    ) -> Union[float, "Datum"]:  # lgtm [py/similar-function]
        """
        Access a covalent radius for species `atom`.

        Parameters
        ----------
        atom : int or str
            Identifier for element or nuclide, e.g., ``H``, ``D``, ``H2``, ``He``, ``hE4``.
            In general, one value recommended for each element; however, certain other exact labels may be available.
            ALVAREZ2008: C_sp3, C_sp2, C_sp, Mn_lowspin, Mn_highspin, Fe_lowspin, Fe_highspin, Co_lowspin, Co_highspin
        units : str, optional
            Units of returned value. To return in native unit (ALVAREZ2008: angstrom), pass it explicitly.
            Only relevant for ``return_tuple=False`` since ``True`` returns underlying data structure with native units.
        missing : float or None, optional
            How to handle when ``atom`` is valid but outside the available data range. When ``None``, raises DataUnavailableError.
            When a float, returns that float, so supply in ``units`` units. Supplying a float is a more compact assurance
            that a call will work over all the periodic table than the equivalent

            .. code-block:: python

                try:
                    rad = qcel.covalentradii.get(atom)
                except qcel.DataUnavailableError:
                    rad = 4.0

            Only relevant for ``return_tuple=False``.
        return_tuple : bool, optional
            See below.

        Returns
        -------
        float
            When ``return_tuple=False``, value of covalent radius. If multiple defined for element, returns largest.
        qcelemental.Datum
            When ``return_tuple=True``, Datum with units, description, uncertainty, and value of covalent radius as Decimal (preserving significant figures).
            If multiple defined for element, returns largest.

        Raises
        ------
        NotAnElementError
            If `atom` cannot be resolved into an element or nuclide or label.
        DataUnavailableError
            If `atom` is a valid element or nuclide but not one for which a covalent radius is available and `missing=None`.

        """
        if atom in self.cr.keys():
            # catch extra labels like 'C_sp3'
            identifier = atom
        else:
            identifier = periodictable.to_E(atom)

        try:
            assert isinstance(identifier, str)  # Should be string by now
            qca = self.cr[identifier]
        except KeyError as e:
            if missing is not None and return_tuple is False:
                return missing
            else:
                raise DataUnavailableError("covalent radius", identifier) from e

        if return_tuple:
            return qca
        else:
            return qca.to_units(units)

    def string_representation(self) -> str:
        """Print name, value, and units of all covalent radii."""

        return print_variables(self.cr)

    def write_c_header(self, filename: str = "covrad.h", missing: float = 2.0) -> None:  # lgtm[py/similar-function]
        """Write C header file defining covalent radii array.

        Parameters
        ----------
        filename : str, optional
            File name for header. Note that changing this won't change the header guard.
        missing : float, optional
            In order that the C array be atomic-number indexable and that it span the
            periodic table, this value is used anywhere data is missing.

        """
        text = [
            "#ifndef _qcelemental_covrad_h_",
            "#define _qcelemental_covrad_h_",
            "",
            "/* This file is autogenerated from the QCElemental python module */",
            "",
            "const double covalent_radii[] = {",
        ]

        for el in periodictable.E:
            try:
                qca = self.cr[el]
                text.append("{},  /*- [{}] {} {} -*/".format(qca.data, qca.units, qca.label, qca.comment))
            except KeyError:
                text.append(
                    "{:.2f},  /*- [{}] {} {} -*/".format(
                        missing, self.native_units, el, "Default value for missing data"
                    )
                )

        text.append("};")
        text.append("#endif /* header guard */")
        text.append("")

        with open(filename, "w") as handle:
            handle.write("\n".join(text))
        print("File written ({}). Remember to add license and clang-format it.".format(filename))


# singleton
covalentradii = CovalentRadii("ALVAREZ2008")



import itertools
from typing import Any, Dict, List, Union

import numpy as np

from ..exceptions import ValidationError
from ..util import unique_everseen


def _apply_default(llist, default):
    return [default if (c is None) else c for c in llist]


def _high_spin_sum(mult_list):
    mm = 1
    for m in mult_list:
        mm += m - 1
    return mm


def _mult_ok(m):
    return isinstance(m, (int, np.integer)) and m >= 1


def _sufficient_electrons_for_mult(z, c, m):
    """Require sufficient electrons in total: total mult ({}) - 1 > raw electrons ({}) - total chg ({})"""
    return m - 1 <= z - c


def _parity_ok(z, c, m):
    """Check total electrons (neutral protons `z` and charge `c`) is (un)paired-compatible with multiplicity `m`"""
    return (m % 2) != ((z - c) % 2)


# def _alpha_beta_allocator(z, c, m):
#    nbeta = (z - c - m + 1) // 2
#    nalpha = nbeta + m - 1
#    return nalpha, nbeta


def validate_and_fill_chgmult(
    zeff,
    fragment_separators,
    molecular_charge: Union[float, None],
    fragment_charges,
    molecular_multiplicity: Union[int, None],
    fragment_multiplicities,
    zero_ghost_fragments: bool = False,
    verbose: int = 1,
) -> Dict[str, Any]:
    """Forms molecular and fragment charge and multiplicity specification
    by completing and reconciling information from argument, supplemented
    by physical constraints and sensible defaults.

    Parameters
    ----------
    zeff : ndarray of float
        (nat,) electron counts for neutral atoms, generally Z nuclear charge.
        0 indicates ghosts such that a full fragment of 0s will be constained
        to `0 1` charge & multiplicity.
    fragment_separators : ndarray of int
        (nfr - 1, ) indices splitting `zeff` into nfr fragments.
    molecular_charge : float or None
        Total charge for molecular system.
    fragment_charges : list of float or None
        (nfr,) known fragment charges with `None` as placeholder for
        unknown. Expected pre-defaulted so even if nothing known, if
        `fragment_separators` breaks `zeff` into `nfr=2` fragments, input
        value should be ``fragment_charges=[None, None]``.
    molecular_multiplicity : int or None
        Total multiplicity for molecular system.
    fragment_multiplicities : list of int or None
        (nfr,) known fragment charges with `None` as placeholder for
        unknown. Expected pre-defaulted so even if nothing known, if
        `fragment_separators` breaks `zeff` into `nfr=2` fragments, input
        value should be ``fragment_multiplicities=[None, None]``.
    zero_ghost_fragments : bool, optional
        Fragments composed entirely of ghost atoms (Zeff=0) are required to have
        chgmult `0 1`. When `False`, violations of this will cause a
        ValidationError. When `True`, treat ghost fragments indicated by `zeff` to
        contain superior information over chgmult arguments that might still
        correspond to full-real molecule. Clears information from
        `molecular_charge` and `molecular_multiplicity` and sets ghost fragments
        to `0 1`, leaving other positions free to readjust. Unused (prefer to set
        up such manipulations outside function call) but works.
    verbose : int, optional
        Amount of printing.

    Returns
    -------
    molecular_charge : float
        Total charge for molecular system.
    fragment_charges : list of float
        (nfr,) Charge on each fragment.
    molecular_multiplicity : int
        Total multiplicity for molecular system.
    fragment_multiplicities : list of int
        (nfr,) Multiplicity for each fragment.

    Raises
    ------
    qcelemental.ValidationError
        When no solution to input arguments subject to the constraints below can be found.

    Notes
    -----
    Returns combination of total & fragment charge & multiplicity among
    values of S1-7 that fulfill rules R1-9. A few derived implications in I1-3.

    * Constraints

      * R1 require all chg & mult exist
      * R2 require total charge to be the sum of frag chg
      * R3 require mult is positive int
      * R4 require sufficient tot electrons for mult: mult - 1 <= neut_el - chg
      * R5 require total parity consistent among tot electrons and mult: (mult % 2) != ((neut_el - chg) % 2)
      * R6 require chg match input argument values
      * R7 require mult match input argument values
      * R8 require that tot = sum(frag) mult follow high spin addition unless tot & frag mult fully specified
      * R9 require that ghost fragments (zeff all 0) be neutral singlet

    * Allowed values

      * S1 suggest input argument values for tot chg, frag chg, tot mult or frag mult
      * S2 suggest sum frag chg for tot chg, allowing for indiv frag chg defaulting to 0
      * S3 suggest distributing unallocated chg onto frag chg
      * S4 suggest 0 default for frag chg
      * S5 suggest range of high-spin sum frag mult for tot mult, allowing for indiv frag mult defaulting to 1 or 2
      * S6 suggest range of unallocated mult = tot - high_spin_sum(frag - 1), allowing for all indiv but self defaulting to 1 or 2.
      * S7 suggest 1 or 2 default for frag mult

    * Implications

      * I1 won't form an ion just to be closed shell (would require choosing +1 vs. -1)
      * I2 unallocated chg or mult lands on the first unspecified fragment able to bear it (enforced by returning first match encountered; subsequent matches distribute charge to later frags)
      * I3 missing chg or mult from tot - frags will always be allocated as a block, not distributed

    Examples
    --------
    >>> validate_and_fill_chgmult(*sys('He'), 0, [0], 1, [1])
    0, [0], 1, [1]
    >>> validate_and_fill_chgmult(*sys('He'), None, [None], None, [None])
    0, [0], 1, [1]
    >>> validate_and_fill_chgmult(*sys('He/He'), None, [None, None], None, [None, None])
    0, [0, 0], 1, [1, 1])
    >>> validate_and_fill_chgmult(*sys('He/He'), 2, [None, None], None, [None, None])
    2, [2, 0], 1, [1, 1])
    >>> validate_and_fill_chgmult(*sys('He/He'), None, [2, None], None, [None, None])
    2, [2, 0], 1, [1, 1])
    >>> validate_and_fill_chgmult(*sys('He/He'), 0, [2, None], None, [None, None])
    0, [2, -2], 1, [1, 1])
    >>> validate_and_fill_chgmult(*sys('Ne/He/He'), -2, [None, 2, None], None, [None, None, None])
    -2, [-4, 2, 0], 1, [1, 1, 1]
    >>> validate_and_fill_chgmult(*sys('Ne/He/He'), 2, [None, -2, None], None, [None, None, None])
    2, [4, -2, 0], 1, [1, 1, 1]
    #  9 - residual +4 distributes to first fragment able to wholly accept it (He+4 is no-go)
    >>> validate_and_fill_chgmult(*sys('He/He/Ne'), 2, [None, -2, None], None, [None, None, None])
    2, [0, -2, 4], 1, [1, 1, 1]
    # 10 - residual +4 unsuited for only open fragment, He, so irreconcilable
    >>> validate_and_fill_chgmult(*sys('He/He/Ne'), 2, [None, -2, 0], None, [None, None, None])
    ValidationError
    # 11 - non-positive multiplicity
    >>> validate_and_fill_chgmult(*sys('He/He/Ne'), 2, [2, -2, None], None, [None, None, None])
    2, [2, -2, 2], 1, [1, 1, 1])
    >>> validate_and_fill_chgmult(*sys('He/He'), None, [-2, 2], None, [None, None])
    0, [-2, 2], 1, [1, 1]
    >>> validate_and_fill_chgmult(*sys('He/He'), None, [None, -2], None, [None, None])
    -2, [0, -2], 1, [1, 1]
    >>> validate_and_fill_chgmult(*sys('Ne/Ne'), 0, [None, 4], None, [None, None])
    0, [-4, 4], 1, [1, 1]
    >>> validate_and_fill_chgmult(*sys('He/He/He'), 4, [2, None, None], None, [None, None, None])
    4, [2, 2, 0], 1, [1, 1, 1]
    >>> validate_and_fill_chgmult(*sys('He/He'), 0, [-2, 2], None, [None, None])
    0, [-2, 2], 1, [1, 1]
    >>> validate_and_fill_chgmult(*sys('He/He'), 0, [-2, -2], None, [None, None])
    ValidationError
    >>> validate_and_fill_chgmult(*sys('He'), None, [None], 0, [None])
    ValidationError
    >>> validate_and_fill_chgmult(*sys('He'), None, [None], None, [1])
    0, [0], 1, [1]
    # 20 - doublet non consistent with closed-shell, neutral default charge
    >>> validate_and_fill_chgmult(*sys('He'), None, [None], None, [2])
    ValidationError
    >>> validate_and_fill_chgmult(*sys('He'), None, [None], None, [3])
    0, [0], 3, [3]
    # 22 - insufficient electrons for pentuplet
    >>> validate_and_fill_chgmult(*sys('He'), None, [None], None, [5])
    ValidationError
    >>> validate_and_fill_chgmult(*sys('He'), None, [-1], None, [2])
    -1, [-1], 2, [2]
    # 24 - doublet not consistent with even charge
    >>> validate_and_fill_chgmult(*sys('He'), None, [-2], None, [2])
    ValidationError
    >>> validate_and_fill_chgmult(*sys('He/He'), None, [None, None], None, [1, 1])
    0, [0, 0], 1, [1, 1]
    >>> validate_and_fill_chgmult(*sys('He/He'), None, [None, None], None, [3, 1])
    0, [0, 0], 3, [3, 1]
    >>> validate_and_fill_chgmult(*sys('He/He'), None, [None, None], None, [1, 3])
    0, [0, 0], 3, [1, 3]
    >>> validate_and_fill_chgmult(*sys('He/He'), None, [None, None], None, [3, 3])
    0, [0, 0], 5, [3, 3]
    >>> validate_and_fill_chgmult(*sys('He/He'), None, [None, None], 3, [3, 3])
    0, [0, 0], 3, [3, 3]
    # 30 - bad parity btwn mult and total # electrons
    >>> validate_and_fill_chgmult(*sys('He/He'), None, [None, None], 2, [3, 3])
    ValidationError
    >>> validate_and_fill_chgmult(*sys('H'), None, [None], None, [None])
    0, [0], 2, [2]
    >>> validate_and_fill_chgmult(*sys('H'), 1, [None], None, [None])
    1, [1], 1, [1]
    >>> validate_and_fill_chgmult(*sys('H'), None, [-1], None, [None])
    -1, [-1], 1, [1]
    >>> validate_and_fill_chgmult(*sys('funnyH'), None, [None], None, [None])
    0, [0], 1, [1]
    # 35 - insufficient electrons
    >>> validate_and_fill_chgmult(*sys('funnierH'), None, [None], None, [None])
    ValidationError
    >>> validate_and_fill_chgmult(*sys('H/H'), None, [None, None], None, [None, None])
    0, [0, 0], 3, [2, 2]
    >>> validate_and_fill_chgmult(*sys('H/He'), None, [None, None], None, [None, None])
    0, [0, 0], 2, [2, 1]
    >>> validate_and_fill_chgmult(*sys('H/He'), None, [1, 1], None, [None, None])
    2, [1, 1], 2, [1, 2]
    >>> validate_and_fill_chgmult(*sys('H/He'), -2, [-1, None], None, [None, None])
    -2, [-1, -1], 2, [1, 2]
    >>> validate_and_fill_chgmult(*sys('H/He/Na/Ne'), None, [1, None, 1, None], None, [None, None, None, None])
    2, [1, 0, 1, 0], 1, [1, 1, 1, 1]
    >>> validate_and_fill_chgmult(*sys('H/He/Na/Ne'), None, [-1, None, 1, None], None, [None, None, None, None])
    0, [-1, 0, 1, 0], 1, [1, 1, 1, 1]
    >>> validate_and_fill_chgmult(*sys('H/He/Na/Ne'), 2, [None, None, 1, None], None, [None, None, None, None])
    2, [1, 0, 1, 0], 1, [1, 1, 1, 1]
    >>> validate_and_fill_chgmult(*sys('H/He/Na/Ne'), 3, [None, None, 1, None], None, [None, None, None, None])
    3, [0, 2, 1, 0], 2, [2, 1, 1, 1]
    >>> validate_and_fill_chgmult(*sys('H/He'), None, [1, None], None, [2, None])
    ValidationError
    >>> validate_and_fill_chgmult(*sys('H/He'), None, [None, 0], None, [None, 2])
    ValidationError
    >>> validate_and_fill_chgmult(*sys('H/He'), None, [None, -1], None, [None, 3])
    ValidationError
    >>> validate_and_fill_chgmult(*sys('H/He/Na/Ne'), None, [None, 1, 0, 1], None, [None, None, None, None])
    2, [0, 1, 0, 1], 5, [2, 2, 2, 2]
    >>> validate_and_fill_chgmult(*sys('H/He/Na/Ne'), None, [None, 1, 0, None], None, [None, None, None, None])
    1, [0, 1, 0, 0], 4, [2, 2, 2, 1]
    >>> validate_and_fill_chgmult(*sys('H/He/Na/Ne'), None, [None, 1, 0, None], None, [None, None, 4, None])
    1, [0, 1, 0, 0], 6, [2, 2, 4, 1]
    >>> validate_and_fill_chgmult(*sys('He/He/He'), 0, [None, None, 1], None, [1, None, 2])
    0, [0, -1, 1], 3, [1, 2, 2]
    >>> validate_and_fill_chgmult(*sys('N/N/N'), None, [1, 1, 1], 3, [None, 3, None])
    3, [1, 1, 1], 3, [1, 3, 1]
    >>> validate_and_fill_chgmult(*sys('N/N/N'), None, [1, 1, 1], 3, [None, None, None])
    3, [1, 1, 1], 3, [3, 1, 1]
    >>> validate_and_fill_chgmult(*sys('N/N/N'), None, [None, None, None], 3, [None, None, 2])
    ValidationError
    >>> validate_and_fill_chgmult(*sys('N/N/N'), 1, [None, -1, None], 3, [None, None, 2])
    1, [2, -1, 0], 3, [2, 1, 2]
    # 55 - both (1, (1, 0.0, 0.0), 4, (1, 3, 2)) and (1, (0.0, 0.0, 1), 4, (2, 3, 1)) plausible
    >>> validate_and_fill_chgmult(*sys('N/Ne/N'), 1, [None, None, None], 4, [None, 3, None])
    1, [1, 0, 0], 4, [1, 3, 2]
    >>> validate_and_fill_chgmult(*sys('N/Ne/N'), None, [None, None, 1], 4, [None, 3, None])
    1, [0, 0, 1], 4, [2, 3, 1]
    >>> validate_and_fill_chgmult(*sys('He/He'), None, [-1, 1], None, [None, None])
    0, [-1, 1], 3, [2, 2]
    >>> validate_and_fill_chgmult(*sys('Gh'), 1, [None], None, [None])
    ValidationError
    >>> validate_and_fill_chgmult(*sys('Gh'), -1, [None], None, [None])
    ValidationError
    >>> validate_and_fill_chgmult(*sys('Gh'), None, [None], 3, [None])
    ValidationError
    >>> validate_and_fill_chgmult(*sys('He/Gh'), None, [2, None], None, [None, None])
    2, [2, 0], 1, [1, 1]
    >>> validate_and_fill_chgmult(*sys('Gh/He'), None, [2, None], None, [None, None])
    ValidationError
    >>> validate_and_fill_chgmult(*sys('Gh/He/Ne'), 2, [None, -2, None], None, [None, None, None])
    2, [0, -2, 4], 1, [1, 1, 1]
    >>> validate_and_fill_chgmult(*sys('Gh/He/Gh'), 1, [None, None, None], None, [None, None, None])
    1, [0, 1, 0], 2, [1, 2, 1]
    >>> sys = {
        'He': (np.array([2]), np.array([])),
        'He/He': (np.array([2, 2]), np.array([1])),
        'Ne/He/He': (np.array([10, 2, 2]), np.array([1, 2])),
        'He/He/Ne': (np.array([2, 2, 10]), np.array([1, 2])),
        'Ne/Ne': (np.array([10, 10]), np.array([1])),
        'He/He/He': (np.array([2, 2, 2]), np.array([1, 2])),
        'H': (np.array([1]), np.array([])),
        'funnyH': (np.array([0]), np.array([])),  # has no electrons
        'funnierH': (np.array([-1]), np.array([])),  # has positron
        'H/H': (np.array([1, 1]), np.array([1])),
        'H/He': (np.array([1, 2]), np.array([1])),
        'H/He/Na/Ne': (np.array([1, 2, 11, 10]), np.array([1, 2, 3])),
        'N/N/N': (np.array([7, 7, 7]), np.array([1, 2])),
        'N/Ne/N': (np.array([7, 10, 7]), np.array([1, 2])),
        'He/Gh': (np.array([2, 0]), np.array([1])),
        'Gh/He': (np.array([0, 2]), np.array([1])),
        'Gh': (np.array([0, 0]), np.array([])),
        'Gh/He/Ne': (np.array([0, 0, 2, 10]), np.array([2, 3])),
        'Gh/He/Gh': (np.array([0, 2, 0]), np.array([1, 2]))}

    """
    log_full = verbose >= 2
    log_brief = verbose >= 2  # TODO: Move back to 1
    text = []

    felez = np.split(zeff, fragment_separators)
    nfr = len(felez)
    if log_full:
        text.append("felez: {}".format(felez))

    cgmp_exact_c = []  # exact_* are candidates for the final value
    cgmp_exact_fc: List[List[float]] = [[] for f in range(nfr)]
    cgmp_exact_m = []
    cgmp_exact_fm: List[List[int]] = [[] for f in range(nfr)]

    cgmp_range = []  # tests that the final value must pass to be valid
    cgmp_rules = []  # key to what rules in cgmp_range are T/F

    real_fragments = np.array([not all(f == 0 for f in felez[ifr]) for ifr in range(nfr)])
    all_fc_known = all(f is not None for f in fragment_charges)
    all_fm_known = all(f is not None for f in fragment_multiplicities)
    if log_full:
        text.append("all_fc_known: {}".format(all_fc_known))
        text.append("all_fm_known: {}".format(all_fm_known))

    if zero_ghost_fragments and not all(real_fragments):
        if log_brief:
            print("possibly adjusting charges")
        molecular_charge = None
        fragment_charges = [(fr if real_fragments[ifr] else 0.0) for ifr, fr in enumerate(fragment_charges)]
        molecular_multiplicity = None
        fragment_multiplicities = [(fr if real_fragments[ifr] else 1) for ifr, fr in enumerate(fragment_multiplicities)]

    # <<< assert broad physical requirements

    #   * (R1) require all chg & mult exist
    cgmp_range.append(
        lambda c, fc, m, fm: c is not None
        and all(f is not None for f in fc)
        and m is not None
        and all(f is not None for f in fm)
    )
    cgmp_rules.append("1")

    #   * (R2) require total charge to be the sum of fragment charges
    cgmp_range.append(lambda c, fc, m, fm: c == sum(fc))
    cgmp_rules.append("2")

    #   * (R3) require mult is positive int
    cgmp_range.append(lambda c, fc, m, fm: _mult_ok(m) and all(_mult_ok(f) for f in fm))
    cgmp_rules.append("3")

    # <<< assert electron count requirements

    zel = np.sum(zeff)  # note: number electrons in neutral species, not number total electrons
    fzel = [np.sum(f) for f in felez]
    if log_full:
        text.append("zel: {}".format(zel))
        text.append("fzel: {}".format(fzel))

    #   * (R4) require sufficient electrons for mult: mult - 1 <= neutral_electrons - chg
    cgmp_range.append(lambda c, fc, m, fm: _sufficient_electrons_for_mult(zel, c, m))
    cgmp_rules.append("4")
    for ifr in range(nfr):
        cgmp_range.append(
            lambda c, fc, m, fm, ifr=ifr: _sufficient_electrons_for_mult(fzel[ifr], fc[ifr], fm[ifr])
        )  # type: ignore
        cgmp_rules.append("4-" + str(ifr))

    #   * (R5) require total parity consistent among neutral_electrons, chg, and mult
    cgmp_range.append(lambda c, fc, m, fm: _parity_ok(zel, c, m))
    cgmp_rules.append("5")
    for ifr in range(nfr):
        cgmp_range.append(lambda c, fc, m, fm, ifr=ifr: _parity_ok(fzel[ifr], fc[ifr], fm[ifr]))  # type: ignore
        cgmp_rules.append("5-" + str(ifr))

    # <<< (R6, R7, S1) assert & suggest input values

    if molecular_charge is not None:
        cgmp_exact_c.append(molecular_charge)
        cgmp_range.append(lambda c, fc, m, fm: c == molecular_charge)
        cgmp_rules.append("6")
    for ifr, chg in enumerate(fragment_charges):
        if chg is not None:
            cgmp_exact_fc[ifr].append(chg)
            cgmp_range.append(lambda c, fc, m, fm, ifr=ifr, chg=chg: fc[ifr] == chg)  # type: ignore
            cgmp_rules.append("6-" + str(ifr))
    if molecular_multiplicity is not None:
        cgmp_exact_m.append(molecular_multiplicity)
        cgmp_range.append(lambda c, fc, m, fm: m == molecular_multiplicity)
        cgmp_rules.append("7")
    for ifr, mult in enumerate(fragment_multiplicities):
        if mult is not None:
            cgmp_exact_fm[ifr].append(mult)
            cgmp_range.append(lambda c, fc, m, fm, ifr=ifr, mult=mult: fm[ifr] == mult)  # type: ignore
            cgmp_rules.append("7-" + str(ifr))

    # <<< assert high-spin-rule and suggest "missing quantity" and default values

    #   * (S2) suggest net frag charge for total charge, allowing for indiv frag defaulting to 0
    cgmp_exact_c.append(sum(filter(None, fragment_charges)))

    missing_frag_chg = 0.0 if molecular_charge is None else molecular_charge
    missing_frag_chg -= sum(filter(None, fragment_charges))

    #   * (S3) suggest distributing unallocated charge onto fragment
    #   * (S4) suggest 0 default charge for fragment
    for ifr in range(nfr):
        if fragment_charges[ifr] is None:  # unneeded, but shortens the exact lists
            cgmp_exact_fc[ifr].append(missing_frag_chg)
            cgmp_exact_fc[ifr].append(0.0)

    #   * (R8) require that frag mult follow high spin addition unless fully specified
    if molecular_multiplicity is None or any(f is None for f in fragment_multiplicities):
        cgmp_range.append(lambda c, fc, m, fm: m == _high_spin_sum(fm))
        cgmp_rules.append("8")

    #   * (S5) suggest range of net frag mult for total mult, allowing for indiv frag defaulting to 1 or 2.
    #          many in range may be unphysical, but those will be caught by physical rules.
    if molecular_multiplicity is None:  # unneeded, but shortens the exact lists
        frag_mult_hi = _high_spin_sum(_apply_default(fragment_multiplicities, 2))
        frag_mult_lo = _high_spin_sum(_apply_default(fragment_multiplicities, 1))
        for m in range(frag_mult_lo, frag_mult_hi + 1):
            cgmp_exact_m.append(m)

    #   * (S6) suggest range of missing mult = tot - high_spin_sum(frag - 1),
    #          allowing for all indiv but self defaulting to 1 or 2.  Many in range
    #          may be unphysical, but those will be caught by physical rules.
    #   * (S7) suggest 1 or 2 default multiplicity for fragment
    if molecular_multiplicity is not None and any(f is None for f in fragment_multiplicities):
        frag_mult_less_one_none = fragment_multiplicities[:]
        frag_mult_less_one_none.remove(None)  # "missing" slot to solve for
        frag_mult_hi = _high_spin_sum(_apply_default(frag_mult_less_one_none, 2))
        frag_mult_lo = _high_spin_sum(_apply_default(frag_mult_less_one_none, 1))
        missing_mult_hi = molecular_multiplicity - frag_mult_lo + 1
        missing_mult_lo = molecular_multiplicity - frag_mult_hi + 1
    else:
        missing_mult_hi = 0
        missing_mult_lo = 0

    for ifr in range(nfr):
        if fragment_multiplicities[ifr] is None:  # unneeded, but shortens the exact lists
            for m in reversed(range(max(missing_mult_lo, 1), missing_mult_hi + 1)):
                cgmp_exact_fm[ifr].append(m)
            cgmp_exact_fm[ifr].append(1)
            cgmp_exact_fm[ifr].append(2)

    #   * (R9) require that ghost fragments be neutral singlet
    for ifr in range(nfr):
        if all(f == 0 for f in felez[ifr]):
            cgmp_range.append(lambda c, fc, m, fm, ifr=ifr: fc[ifr] == 0 and fm[ifr] == 1)  # type: ignore
            cgmp_rules.append("9-" + str(ifr))

    # <<< reconcile and report

    def reconcile(exact_c, exact_fc, exact_m, exact_fm):
        """Returns a member from all combinations of `exact` that passes all tests in cgmp_range, else raises error."""

        # remove duplicates
        uniq_c = unique_everseen(exact_c)
        uniq_fc = [unique_everseen(f) for f in exact_fc]
        uniq_m = unique_everseen(exact_m)
        uniq_fm = [unique_everseen(f) for f in exact_fm]
        text.append("c: {}".format(list(exact_c)))
        for f in exact_fc:
            text.append("fc: {}".format(list(f)))
        text.append("m: {}".format(list(exact_m)))
        for f in exact_fm:
            text.append("fm: {}".format(list(f)))

        header = True
        for candidate in itertools.product(*[uniq_c, itertools.product(*uniq_fc), uniq_m, itertools.product(*uniq_fm)]):
            cc, cfc, cm, cfm = candidate
            if header:
                if log_full:
                    text.append(
                        """Assess candidate {}: {}""".format(
                            candidate, " ".join(("{:3}".format(r) for r in cgmp_rules))
                        )
                    )
                header = False
            assessment = [fn(cc, cfc, cm, cfm) for fn in cgmp_range]
            sass = ["{:3}".format("T" if b else "") for b in assessment]
            if log_full:
                text.append("""Assess candidate {:}: {} --> {}""".format(candidate, " ".join(sass), all(assessment)))
            if all(assessment):
                return candidate

        err = """Inconsistent or unspecified chg/mult: sys chg: {}, frag chg: {}, sys mult: {}, frag mult: {}""".format(
            molecular_charge, fragment_charges, molecular_multiplicity, fragment_multiplicities
        )
        if verbose > -1:
            print("\n\n" + "\n".join(text))
        raise ValidationError(err)

    def stringify(start, final):
        fcgmp = "{:^4}"
        return fcgmp.format(final) if final == start else fcgmp.format("(" + str(int(final)) + ")")

    # TODO could winnow down the exact_* lists a bit by ruling out
    #      independent values. do this if many-frag molecular systems take too
    #      long in the itertools.product

    c_final, fc_final, m_final, fm_final = reconcile(cgmp_exact_c, cgmp_exact_fc, cgmp_exact_m, cgmp_exact_fm)

    c_text = stringify(molecular_charge, c_final)
    fc_text = ", ".join((stringify(fs, ff) for fs, ff in zip(fragment_charges, fc_final)))
    m_text = stringify(molecular_multiplicity, m_final)
    fm_text = ", ".join((stringify(fs, ff) for fs, ff in zip(fragment_multiplicities, fm_final)))

    brief = []
    if log_brief:
        brief.append("    {:26} {}".format("      charge = " + c_text, "fragments = " + fc_text))
        brief.append("    {:26} {}".format("multiplicity = " + m_text, "fragments = " + fm_text))

    been_defaulted = []
    if c_text.count("(") + fc_text.count("(") > 1:
        been_defaulted.append("charge")
    if "(" in m_text or "(" in fm_text:
        been_defaulted.append("multiplicity")

    if been_defaulted and log_brief:
        brief.append(
            "    Note: Default values have been applied for {}. Specify intentions in molecule input block".format(
                " and ".join(been_defaulted)
            )
        )

    if (m_final != _high_spin_sum(fm_final)) and log_brief:
        brief.append(
            "    Warning: Total multiplicity is not high-spin sum of fragments; may be clobbered by psi4.core.Molecule.update_geometry()."
        )

    if log_full:
        print("\n".join(text))
    if log_brief:
        # TODO add back when printing worked out
        # print('\n'.join(brief))
        pass

    return {
        "molecular_charge": float(c_final),
        "fragment_charges": [float(f) for f in fc_final],
        "molecular_multiplicity": m_final,
        "fragment_multiplicities": list(fm_final),
    }



from typing import List, Optional, Tuple, Union

import numpy as np

from ..covalent_radii import covalentradii
from ..exceptions import NotAnElementError

__all__ = ["guess_connectivity"]


def guess_connectivity(
    symbols: np.ndarray, geometry: np.ndarray, threshold: float = 1.2, default_connectivity: Optional[float] = None
) -> List[Union[Tuple[int, int], Tuple[int, int, float]]]:
    """
    Finds connected atoms based off of a covalent radii metric.

    Parameters
    ----------
    symbols : np.ndarray
        The molecular symbols (e.g., 'Zr', 'C')
    geometry : np.ndarray
        The molecular geometry in Bohr
    threshold : float, optional
        Tunes the covalent radii metric safety factor.
    default_connectivity : Optional[float], optional
        Provides a default connectivity value

    Returns
    -------
    List[Union[Tuple[int, int], Tuple[int, int, float]]]
        Provides a list of connected atoms, or optionally a list of
        connected atoms and default connectivity if provided.
    """

    geometry = np.asarray(geometry, dtype=float).reshape(-1, 3)
    radii = []
    for s in symbols:
        try:
            radii.append(covalentradii.get(s, missing=1.8))
        except NotAnElementError:
            radii.append(1.8)

    radii = np.array(radii)

    # Upper triangular
    con = []
    for x in range(geometry.shape[0]):
        diffs = geometry[x] - geometry[x + 1 :]
        dists = np.einsum("ij,ij->i", diffs, diffs)
        np.sqrt(dists, out=dists)

        cutoff = (radii[x] + radii[x + 1 :]) * threshold
        where = np.where(dists < cutoff)[0]
        where += x + 1

        for atom2 in where:
            con.append((x, atom2))

    if default_connectivity:
        con = [(x[0], x[1], default_connectivity) for x in con]

    return con



import os
from decimal import Decimal

import pytest

import qcelemental
from qcelemental.exceptions import NotAnElementError


@pytest.mark.parametrize(
    "inp,expected",
    [("He", "He"), ("heliuM", "He"), ("He4", "He4"), ("he", "He"), ("2", "He"), (2, "He"), (2.0, "He"), ("D", "D")],
)
def test_id_resolution(inp, expected):
    assert qcelemental.periodictable._resolve_atom_to_key(inp) == expected


@pytest.mark.parametrize("inp", ["He100", "-1", -1, -1.0, "cat", 200])
def test_id_resolution_error(inp):
    with pytest.raises(NotAnElementError):
        qcelemental.periodictable._resolve_atom_to_key(inp)


@pytest.mark.parametrize(
    "inp,expected", [("He", "He"), ("heliuM", "He"), ("he", "He"), ("2", "He"), (2, "He"), (2.0, "He")]
)
def test_id_resolution_strict(inp, expected):
    assert qcelemental.periodictable._resolve_atom_to_key(inp, strict=True) == expected


@pytest.mark.parametrize("inp", ["He100", "-1", -1, -1.0, "cat", 200, "He4", "D"])
def test_id_resolution_strict_error(inp):
    with pytest.raises(NotAnElementError):
        qcelemental.periodictable._resolve_atom_to_key(inp, strict=True)


# TODO test ghost


def test_to_mass_krypton_decimal():
    assert qcelemental.periodictable.to_mass("kr", return_decimal=True) == Decimal("83.9114977282")


@pytest.mark.parametrize(
    "inp,expected",
    [
        # Kr 84
        ("KRYPTON", 83.9114977282),
        ("kr", 83.9114977282),
        ("kr84", 83.9114977282),
        (36, 83.9114977282),
        # Kr 86
        ("kr86", 85.9106106269),
        # Helium
        ("D", 2.01410177812),
        ("h2", 2.01410177812),
    ],
)
def test_to_mass(inp, expected):
    assert qcelemental.periodictable.to_mass(inp) == pytest.approx(expected, 1.0e-9)


@pytest.mark.parametrize(
    "inp,expected",
    [
        # Kr 84
        ("kr", 84),
        ("KRYPTON", 84),
        ("kr84", 84),
        (36, 84),
        # Kr 86
        ("kr86", 86),
        # Helium
        ("D", 2),
        ("h2", 2),
    ],
)
def test_to_mass_number(inp, expected):
    assert qcelemental.periodictable.to_A(inp) == expected
    assert qcelemental.periodictable.to_mass_number(inp) == expected


@pytest.mark.parametrize(
    "inp,expected",
    [
        # Kr 84
        ("kr", 36),
        ("KRYPTON", 36),
        ("kr84", 36),
        (36, 36),
        # Kr 86
        ("kr86", 36),
        # Deuterium
        ("D", 1),
        ("h2", 1),
    ],
)
def test_to_atomic_number(inp, expected):
    assert qcelemental.periodictable.to_Z(inp) == expected
    assert qcelemental.periodictable.to_atomic_number(inp) == expected


@pytest.mark.parametrize(
    "inp,expected",
    [
        # Kr 84
        ("kr", 36),
        ("KRYPTON", 36),
        ("kr84", None),
        (36, 36),
        # Kr 86
        ("kr86", None),
        # Deuterium
        ("D", None),
        ("h2", None),
    ],
)
def test_to_atomic_number_strict(inp, expected):
    if expected is None:
        with pytest.raises(NotAnElementError):
            qcelemental.periodictable.to_Z(inp, strict=True)
        with pytest.raises(NotAnElementError):
            qcelemental.periodictable.to_atomic_number(inp, strict=True)
    else:
        assert qcelemental.periodictable.to_Z(inp, strict=True) == expected
        assert qcelemental.periodictable.to_atomic_number(inp, strict=True) == expected


@pytest.mark.parametrize(
    "inp,expected",
    [
        # Kr 84
        ("kr", "Kr"),
        ("KRYPTON", "Kr"),
        ("kr84", "Kr"),
        (36, "Kr"),
        # Kr 86
        ("kr86", "Kr"),
        # Deuterium
        ("D", "H"),
        ("h2", "H"),
    ],
)
def test_to_symbol(inp, expected):
    assert qcelemental.periodictable.to_E(inp) == expected
    assert qcelemental.periodictable.to_symbol(inp) == expected


@pytest.mark.parametrize(
    "inp,expected",
    [
        # Kr 84
        ("kr", "Kr"),
        ("KRYPTON", "Kr"),
        ("kr84", None),
        (36, "Kr"),
        # Kr 86
        ("kr86", None),
        # Deuterium
        ("D", None),
        ("h2", None),
    ],
)
def test_to_symbol_strict(inp, expected):
    if expected is None:
        with pytest.raises(NotAnElementError):
            qcelemental.periodictable.to_E(inp, strict=True)
        with pytest.raises(NotAnElementError):
            qcelemental.periodictable.to_symbol(inp, strict=True)
    else:
        assert qcelemental.periodictable.to_E(inp, strict=True) == expected
        assert qcelemental.periodictable.to_symbol(inp, strict=True) == expected


@pytest.mark.parametrize(
    "inp,expected",
    [
        # Kr 84
        ("kr", "Krypton"),
        ("KRYPTON", "Krypton"),
        ("kr84", "Krypton"),
        (36, "Krypton"),
        # Kr 86
        ("kr86", "Krypton"),
        # Deuterium
        ("D", "Hydrogen"),
        ("h2", "Hydrogen"),
    ],
)
def test_to_element(inp, expected):
    assert qcelemental.periodictable.to_element(inp) == expected
    assert qcelemental.periodictable.to_name(inp) == expected


@pytest.mark.parametrize(
    "inp,expected",
    [
        # Kr 84
        ("kr", "Krypton"),
        ("KRYPTON", "Krypton"),
        ("kr84", None),
        (36, "Krypton"),
        # Kr 86
        ("kr86", None),
        # Deuterium
        ("D", None),
        ("h2", None),
    ],
)
def test_to_element_strict(inp, expected):
    if expected is None:
        with pytest.raises(NotAnElementError):
            qcelemental.periodictable.to_element(inp, strict=True)
        with pytest.raises(NotAnElementError):
            qcelemental.periodictable.to_name(inp, strict=True)
    else:
        assert qcelemental.periodictable.to_element(inp, strict=True) == expected
        assert qcelemental.periodictable.to_name(inp, strict=True) == expected


@pytest.mark.parametrize(
    "inp,expected", [("HE", 1), ("carbon", 2), ("cl35", 3), (36, 4), (37.0, 5), ("mercury", 6), ("Ts", 7)]
)
def test_to_period(inp, expected):
    assert qcelemental.periodictable.to_period(inp) == expected


@pytest.mark.parametrize(
    "inp,expected",
    [
        ("D", 1),
        ("Ba", 2),
        ("yttrium", 3),
        ("hf179", 4),
        (23, 5),
        ("w", 6),
        (43.0, 7),
        (26.0, 8),
        ("meitnerium", 9),
        ("pt", 10),
        ("gold", 11),
        ("cd", 12),
        (5.0, 13),
        (14, 14),
        ("phosphorus", 15),
        ("SE", 16),
        ("i", 17),
        (54, 18),
        ("La", None),
        (89.0, None),
        ("lutetium", None),
        (103.0, None),
        ("u238", None),
    ],
)
def test_to_group(inp, expected):
    assert qcelemental.periodictable.to_group(inp) == expected


def test_c_header():
    from qcelemental.periodic_table import write_c_header

    write_c_header("header.h")
    os.remove("header.h")


@pytest.mark.xfail(True, reason="comparison data not available for installed repository", run=True, strict=False)
def test_periodic_table_comparison():
    from qcelemental.periodic_table import run_comparison

    run_comparison()
